# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

# pylint: disable=too-many-lines

"""
Test the project functionality
"""


import unittest
from pathlib import Path
import os
from shutil import rmtree
from time import sleep
import itertools
from unittest import mock
from vunit.exceptions import CompileError
from vunit.ostools import renew_path, write_file
from vunit.project import Project
from vunit.source_file import file_type_of


class TestProject(unittest.TestCase):  # pylint: disable=too-many-public-methods
    """
    Test the Project class
    """

    def setUp(self):
        self.output_path = str(Path(__file__).parent / "test_project_out")
        renew_path(self.output_path)
        self.project = Project()
        self.cwd = os.getcwd()
        os.chdir(self.output_path)

    def tearDown(self):
        os.chdir(self.cwd)
        if Path(self.output_path).exists():
            rmtree(self.output_path)

    def test_parses_entity_architecture(self):
        self.project.add_library("lib", "work_path")

        # Add architecture before entity to test that they are paired later
        self.add_source_file(
            "lib",
            "file2.vhd",
            """\
architecture arch3 of foo is
begin
end architecture;
""",
        )

        file1 = self.add_source_file(
            "lib",
            "file1.vhd",
            """\
entity foo is
end entity;

architecture arch of foo is
begin
end architecture;

architecture arch2 of foo is
begin
end architecture;
""",
        )

        self.assert_has_entity(file1, "foo", architecture_names=["arch", "arch2", "arch3"])
        self.add_source_file(
            "lib",
            "file3.vhd",
            """\
architecture arch4 of foo is
begin
end architecture;
""",
        )

        self.assert_has_entity(file1, "foo", architecture_names=["arch", "arch2", "arch3", "arch4"])
        self.assert_has_architecture("file1.vhd", "arch", "foo")
        self.assert_has_architecture("file1.vhd", "arch2", "foo")
        self.assert_has_architecture("file2.vhd", "arch3", "foo")
        self.assert_has_architecture("file3.vhd", "arch4", "foo")

    def test_parses_entity_architecture_with_generics(self):
        self.project.add_library("lib", "work_path")
        file1 = self.add_source_file(
            "lib",
            "file1.vhd",
            """\
entity foo is
  generic (
    testing_that_foo : boolean;
    testing_that_bar : boolean);
end entity;

architecture arch of foo is
begin
end architecture;
""",
        )

        self.assert_has_entity(
            file1,
            "foo",
            generic_names=["testing_that_bar", "testing_that_foo"],
            architecture_names=["arch"],
        )
        self.assert_has_architecture("file1.vhd", "arch", "foo")

    def test_parses_package(self):
        self.project.add_library("lib", "work_path")
        self.add_source_file(
            "lib",
            "file1.vhd",
            """\
package foo is
end package;

package body foo is
begin
end package body;
""",
        )
        self.assert_has_package("file1.vhd", "foo")
        self.assert_has_package_body("file1.vhd", "foo")

    @mock.patch("vunit.source_file.LOGGER")
    def test_recovers_from_parse_error(self, logger):
        self.project.add_library("lib", "work_path")
        source_file = self.add_source_file(
            "lib",
            "file.vhd",
            """\
entity foo is
 port (;);
end entity;
""",
        )
        logger.error.assert_called_once_with("Failed to parse %s", "file.vhd")
        self.assertEqual(source_file.design_units, [])

    def test_finds_entity_instantiation_dependencies(self):
        file1, file2, file3 = self.create_dummy_three_file_project()

        self.assert_compiles(file1, before=file2)
        self.assert_compiles(file2, before=file3)

    def test_primary_with_same_name_in_multiple_libraries_secondary_dependency(self):
        self.project.add_library("lib1", "lib1_path")
        self.project.add_library("lib2", "lib2_path")

        foo_arch = self.add_source_file(
            "lib1",
            "foo_arch.vhd",
            """
architecture arch of foo is
begin
end architecture;
""",
        )

        foo1_ent = self.add_source_file(
            "lib1",
            "foo1_ent.vhd",
            """
entity foo is
port (signal bar : boolean);
end entity;
""",
        )

        self.add_source_file(
            "lib2",
            "foo2_ent.vhd",
            """
entity foo is
end entity;
""",
        )

        self.assert_compiles(foo1_ent, before=foo_arch)

    def test_multiple_identical_file_names_with_different_path_in_same_library(self):
        self.project.add_library("lib", "lib_path")
        a_foo = self.add_source_file(
            "lib",
            str(Path("a") / "foo.vhd"),
            """
entity a_foo is
end entity;
""",
        )

        b_foo = self.add_source_file(
            "lib",
            str(Path("b") / "foo.vhd"),
            """
entity b_foo is
end entity;
""",
        )
        self.assert_should_recompile([a_foo, b_foo])
        self.update(a_foo)
        self.assert_should_recompile([b_foo])
        self.update(b_foo)
        self.assert_should_recompile([])

    def test_finds_entity_architecture_dependencies(self):
        self.project.add_library("lib", "lib_path")
        entity = self.add_source_file(
            "lib",
            "entity.vhd",
            """
entity foo is
end entity;
""",
        )

        arch1 = self.add_source_file(
            "lib",
            "arch1.vhd",
            """
architecture arch1 of foo is
begin
end architecture;
""",
        )

        arch2 = self.add_source_file(
            "lib",
            "arch2.vhd",
            """
architecture arch2 of foo is
begin
end architecture;
""",
        )
        self.assert_compiles(entity, before=arch1)
        self.assert_compiles(entity, before=arch2)

    def test_finds_package_dependencies(self):
        self.project.add_library("lib", "lib_path")
        package = self.add_source_file(
            "lib",
            "package.vhd",
            """
package foo is
end package;
""",
        )

        body = self.add_source_file(
            "lib",
            "body.vhd",
            """
package body foo is
begin
end package body;
""",
        )

        self.assert_compiles(package, before=body)

    def create_module_package_and_body(self, add_body=True):
        """
        Help function to create a three file project
        with a package, a package body and a module using the package
        """
        self.project.add_library("lib", "lib_path")

        package = self.add_source_file(
            "lib",
            "package.vhd",
            """
package pkg is
end package;
""",
        )

        body = None
        if add_body:
            body = self.add_source_file(
                "lib",
                "body.vhd",
                """
package body pkg is
begin
end package body;
""",
            )

        self.project.add_library("lib2", "work_path")
        module = self.add_source_file(
            "lib2",
            "module.vhd",
            """
library lib;
use lib.pkg.all;

entity module is
end entity;

architecture arch of module is
begin
end architecture;
""",
        )
        return package, body, module

    def test_finds_use_package_dependencies_case_insensitive(self):
        for library_clause, use_clause in itertools.combinations(("lib", "Lib"), 2):
            self.project = Project()
            self.project.add_library("Lib", "lib_path")

            package = self.add_source_file(
                "Lib",
                "package.vhd",
                """
package pkg is
end package;

package body PKG is
begin
end package body;
""",
            )

            self.project.add_library("lib2", "lib2_path")
            module = self.add_source_file(
                "lib2",
                "module.vhd",
                """
library {library_clause};
use {use_clause}.PKG.all;
            """.format(
                    library_clause=library_clause, use_clause=use_clause
                ),
            )
            self.assert_compiles(package, before=module)

    def test_error_on_case_insensitive_library_name_conflict(self):
        self.project.add_library("Lib", "lib_path1")
        try:
            self.project.add_library("lib", "lib_path1")
        except RuntimeError as exception:
            self.assertEqual(
                str(exception),
                "Library name 'lib' not case-insensitive unique. " "Library name 'Lib' previously defined",
            )
        else:
            raise AssertionError("RuntimeError not raised")

    def test_finds_use_package_dependencies(self):
        package, body, module = self.create_module_package_and_body()
        self.assert_compiles(package, before=body)
        self.assert_compiles(package, before=module)
        self.assert_not_compiles(body, before=module)

    def test_finds_extra_package_body_dependencies(self):
        self.project = Project(depend_on_package_body=True)
        package, body, module = self.create_module_package_and_body()
        self.assert_compiles(package, before=body)
        self.assert_compiles(body, before=module)
        self.assert_compiles(package, before=module)

    def test_that_package_can_have_no_body(self):
        self.project = Project(depend_on_package_body=True)
        package, _, module = self.create_module_package_and_body(add_body=False)
        self.assert_compiles(package, before=module)

    def test_package_instantiation_dependencies_on_generic_package(self):
        self.project.add_library("pkg_lib", "pkg_lib_path")
        pkg = self.add_source_file(
            "pkg_lib",
            "pkg.vhd",
            """
package pkg is
end package;
        """,
        )

        self.project.add_library("lib", "lib_path")
        ent = self.add_source_file(
            "lib",
            "ent.vhd",
            """
library pkg_lib;

entity ent is
end entity;

architecture a of ent is
   package pkg_inst is new pkg_lib.pkg;
begin
end architecture;
""",
        )

        self.assert_compiles(pkg, before=ent)

    def test_package_instantiation_dependencies_on_instantiated_package(self):
        self.project.add_library("lib", "lib_path")
        generic_pkg = self.add_source_file(
            "lib",
            "generic_pkg.vhd",
            """
package generic_pkg is
  generic (foo : boolean);
end package;
""",
        )
        instance_pkg = self.add_source_file(
            "lib",
            "instance_pkg.vhd",
            """
package instance_pkg is new work.generic_pkg
  generic map (foo => false);
""",
        )
        user = self.add_source_file(
            "lib",
            "user.vhd",
            """
use work.instance_pkg;
""",
        )
        self.assert_compiles(generic_pkg, before=instance_pkg)
        self.assert_compiles(instance_pkg, before=user)

    def test_finds_context_dependencies(self):
        self.project.add_library("lib", "lib_path")
        context = self.add_source_file(
            "lib",
            "context.vhd",
            """
context foo is
end context;
""",
        )

        self.project.add_library("lib2", "work_path")
        module = self.add_source_file(
            "lib2",
            "module.vhd",
            """
library lib;
context lib.foo;

entity module is
end entity;

architecture arch of module is
begin
end architecture;
""",
        )

        self.assert_compiles(context, before=module)

    def test_finds_configuration_dependencies(self):
        self.project.add_library("lib", "lib_path")
        cfg = self.add_source_file(
            "lib",
            "cfg.vhd",
            """
configuration cfg of ent is
end configuration;
""",
        )

        ent = self.add_source_file(
            "lib",
            "ent.vhd",
            """
entity ent is
end entity;
""",
        )

        ent_a1 = self.add_source_file(
            "lib",
            "ent_a1.vhd",
            """
architecture a1 of ent is
begin
end architecture;
""",
        )

        ent_a2 = self.add_source_file(
            "lib",
            "ent_a2.vhd",
            """
architecture a2 of ent is
begin
end architecture;
""",
        )

        self.assert_compiles(ent, before=cfg)
        self.assert_compiles(ent_a1, before=cfg)
        self.assert_compiles(ent_a2, before=cfg)

    def test_finds_configuration_reference_dependencies(self):
        self.project.add_library("lib", "lib_path")
        cfg = self.add_source_file(
            "lib",
            "cfg.vhd",
            """
configuration cfg of ent is
end configuration;
""",
        )

        self.add_source_file(
            "lib",
            "ent.vhd",
            """
entity ent is
end entity;
""",
        )

        self.add_source_file(
            "lib",
            "ent_a.vhd",
            """
architecture a of ent is
begin
end architecture;
""",
        )

        top = self.add_source_file(
            "lib",
            "top.vhd",
            """
entity top is
end entity;

architecture a of top is
   for inst : comp use configuration work.cfg;
begin
   inst : comp;
end architecture;
""",
        )

        self.assert_compiles(cfg, before=top)

    def test_specific_architecture_reference_dependencies(self):
        """
        GHDL dependes also on architecture when specificially mentioned
        """
        self.project.add_library("lib", "lib_path")

        self.add_source_file(
            "lib",
            "ent.vhd",
            """
entity ent is
end entity;
""",
        )

        ent_a1 = self.add_source_file(
            "lib",
            "ent_a1.vhd",
            """
architecture a1 of ent is
begin
end architecture;
""",
        )

        ent_a2 = self.add_source_file(
            "lib",
            "ent_a2.vhd",
            """
architecture a2 of ent is
begin
end architecture;
""",
        )

        top1 = self.add_source_file(
            "lib",
            "top1.vhd",
            """
entity top1 is
end entity;

architecture a of top1 is
begin
  inst : entity work.ent(a1);
end architecture;
""",
        )

        top2 = self.add_source_file(
            "lib",
            "top2.vhd",
            """
entity top2 is
end entity;

architecture a of top2 is
  for inst : comp use entity work.ent(a2);
begin
  inst : comp;
end architecture;
""",
        )

        self.assert_compiles(ent_a1, before=top1)
        self.assert_compiles(ent_a2, before=top2)

    def test_work_library_reference_non_lower_case(self):
        """
        Bug discovered in #556
        """
        self.project.add_library("UPPER", "lib_path")

        self.add_source_file(
            "UPPER",
            "ent.vhd",
            """
entity ent is
end entity;
""",
        )

        ent_a1 = self.add_source_file(
            "UPPER",
            "ent_a1.vhd",
            """
architecture a1 of ent is
begin
end architecture;
""",
        )

        top1 = self.add_source_file(
            "UPPER",
            "top1.vhd",
            """
entity top1 is
end entity;

architecture a of top1 is
begin
  inst : entity work.ent(a1);
end architecture;
""",
        )

        self.assert_compiles(ent_a1, before=top1)

    @mock.patch("vunit.project.LOGGER")
    def test_warning_on_missing_specific_architecture_reference(self, mock_logger):
        self.project.add_library("lib", "lib_path")

        self.add_source_file(
            "lib",
            "ent.vhd",
            """
entity ent is
end entity;
""",
        )

        self.add_source_file(
            "lib",
            "arch.vhd",
            """
architecture a1 of ent is
begin
end architecture;
""",
        )

        self.add_source_file(
            "lib",
            "top.vhd",
            """
entity top1 is
end entity;

architecture a of top1 is
begin
  inst1 : entity work.ent(a1);
  inst2 : entity work.ent(a2); # Missing
end architecture;
""",
        )

        self.project.get_files_in_compile_order()
        warning_calls = mock_logger.warning.call_args_list
        log_msg = warning_calls[0][0][0] % warning_calls[0][0][1:]
        self.assertEqual(len(warning_calls), 1)
        self.assertIn("top.vhd", log_msg)
        self.assertIn("a2", log_msg)
        self.assertIn("lib.ent", log_msg)

    @mock.patch("vunit.project.LOGGER")
    def test_error_on_duplicate_file(self, logger):
        self.project.add_library("lib", "lib_path")
        file1 = self.add_source_file("lib", "file.vhd", "entity foo is end entity;")
        self.assertRaises(
            RuntimeError,
            self.add_source_file,
            "lib",
            "file.vhd",
            "entity foo is end entity foo;",
        )

        # No extra design unit of file added
        lib = self.project.get_library("lib")
        self.assertEqual([ent.name for ent in lib.get_entities()], ["foo"])
        self.assertEqual(lib.get_source_file("file.vhd"), file1)
        self.assertEqual(self.project.get_source_files_in_order(), [file1])
        self.assertFalse(logger.warning.called)

    @mock.patch("vunit.library.LOGGER")
    def test_no_error_on_duplicate_identical_file(self, logger):
        self.project.add_library("lib", "lib_path")
        file1 = self.add_source_file("lib", "file.vhd", "entity foo is end entity;")
        file2 = self.add_source_file("lib", "file.vhd", "entity foo is end entity;")
        self.assertEqual(id(file1), id(file2))

        # No extra design unit of file added
        lib = self.project.get_library("lib")
        self.assertEqual([ent.name for ent in lib.get_entities()], ["foo"])
        self.assertEqual(lib.get_source_file("file.vhd"), file1)
        self.assertEqual(self.project.get_source_files_in_order(), [file1])
        self.assertTrue(logger.info.called)

    def _test_warning_on_duplicate(self, code, message, verilog=False):
        """
        Utility function to test adding the same duplicate code under
        file.vhd and file_copy.vhd where the duplication should cause a warning message.
        """
        suffix = "v" if verilog else "vhd"

        self.add_source_file("lib", "file." + suffix, code)

        with mock.patch("vunit.library.LOGGER") as mock_logger:
            self.add_source_file("lib", "file_copy." + suffix, code)
            warning_calls = mock_logger.warning.call_args_list
            log_msg = warning_calls[0][0][0] % warning_calls[0][0][1:]
            self.assertEqual(len(warning_calls), 1)
            self.assertEqual(log_msg, message)

    def test_warning_on_duplicate_entity(self):
        self.project.add_library("lib", "lib_path")
        self._test_warning_on_duplicate(
            """
entity ent is
end entity;
""",
            "file_copy.vhd: entity 'ent' previously defined in file.vhd",
        )

    def test_warning_on_duplicate_package(self):
        self.project.add_library("lib", "lib_path")
        self._test_warning_on_duplicate(
            """
package pkg is
end package;
""",
            "file_copy.vhd: package 'pkg' previously defined in file.vhd",
        )

    def test_warning_on_duplicate_configuration(self):
        self.project.add_library("lib", "lib_path")
        self._test_warning_on_duplicate(
            """
configuration cfg of ent is
end configuration;
""",
            "file_copy.vhd: configuration 'cfg' previously defined in file.vhd",
        )

    def test_warning_on_duplicate_package_body(self):
        self.project.add_library("lib", "lib_path")
        self.add_source_file(
            "lib",
            "pkg.vhd",
            """
package pkg is
end package;
""",
        )

        self._test_warning_on_duplicate(
            """
package body pkg is
end package bodY;
""",
            "file_copy.vhd: package body 'pkg' previously defined in file.vhd",
        )

    def test_warning_on_duplicate_architecture(self):
        self.project.add_library("lib", "lib_path")
        self.add_source_file(
            "lib",
            "ent.vhd",
            """
entity ent is
end entity;
""",
        )

        self.add_source_file(
            "lib",
            "arch.vhd",
            """
architecture a_no_duplicate of ent is
begin
end architecture;
""",
        )

        self._test_warning_on_duplicate(
            """
architecture a of ent is
begin
end architecture;
""",
            "file_copy.vhd: architecture 'a' previously defined in file.vhd",
        )

    def test_warning_on_duplicate_context(self):
        self.project.add_library("lib", "lib_path")
        self._test_warning_on_duplicate(
            """
context ctx is
end context;
""",
            "file_copy.vhd: context 'ctx' previously defined in file.vhd",
        )

    def test_error_on_adding_duplicate_library(self):
        self.project.add_library(logical_name="lib", directory="dir")
        self.assertRaises(ValueError, self.project.add_library, logical_name="lib", directory="dir")

    def test_warning_on_duplicate_verilog_module(self):
        self.project.add_library("lib", "lib_path")
        self._test_warning_on_duplicate(
            """
module foo;
endmodule
""",
            "file_copy.v: module 'foo' previously defined in file.v",
            verilog=True,
        )

    def test_warning_on_duplicate_verilog_package(self):
        self.project.add_library("lib", "lib_path")
        self._test_warning_on_duplicate(
            """
package pkg;
endpackage
""",
            "file_copy.v: package 'pkg' previously defined in file.v",
            verilog=True,
        )

    def test_should_recompile_all_files_initially(self):
        file1, file2, file3 = self.create_dummy_three_file_project()
        self.assert_should_recompile([file1, file2, file3])
        self.assert_should_recompile([file1, file2, file3])

    def test_updating_creates_hash_files(self):
        files = self.create_dummy_three_file_project()

        for source_file in files:
            self.update(source_file)
            self.assertTrue(Path(self.hash_file_name_of(source_file)).exists())

    def test_should_not_recompile_updated_files(self):
        file1, file2, file3 = self.create_dummy_three_file_project()

        self.update(file1)
        self.assert_should_recompile([file2, file3])

        self.update(file2)
        self.assert_should_recompile([file3])

        self.update(file3)
        self.assert_should_recompile([])

    def test_should_recompile_files_affected_by_change(self):
        file1, file2, file3 = self.create_dummy_three_file_project()

        self.update(file1)
        self.update(file2)
        self.update(file3)
        self.assert_should_recompile([])

        file1, file2, file3 = self.create_dummy_three_file_project()
        self.assert_should_recompile([])

        file1, file2, file3 = self.create_dummy_three_file_project(update_file1=True)
        self.assert_should_recompile([file1, file2, file3])

    def test_should_recompile_files_after_changing_compile_options(self):
        file1, file2, file3 = self.create_dummy_three_file_project()

        self.update(file1)
        self.update(file2)
        self.update(file3)
        self.assert_should_recompile([])

        file2.set_compile_option("ghdl.flags", ["--no-vital-checks"])
        self.assert_should_recompile([file2, file3])

    def test_should_recompile_files_after_changing_vhdl_standard(self):
        write_file("file_name.vhd", "")

        self.project = Project()
        self.project.add_library("lib", "lib_path")
        source_file = self.project.add_source_file("file_name.vhd", library_name="lib", vhdl_standard="2008")
        self.assert_should_recompile([source_file])
        self.update(source_file)
        self.assert_should_recompile([])

        self.project = Project()
        self.project.add_library("lib", "lib_path")
        source_file = self.project.add_source_file("file_name.vhd", library_name="lib", vhdl_standard="2002")
        self.assert_should_recompile([source_file])

    def test_add_compile_option(self):
        self.project.add_library("lib", "lib_path")
        file1 = self.add_source_file("lib", "file.vhd", "")
        file1.add_compile_option("ghdl.flags", ["--foo"])
        self.assertEqual(file1.get_compile_option("ghdl.flags"), ["--foo"])
        file1.add_compile_option("ghdl.flags", ["--bar"])
        self.assertEqual(file1.get_compile_option("ghdl.flags"), ["--foo", "--bar"])
        file1.set_compile_option("ghdl.flags", ["--xyz"])
        self.assertEqual(file1.get_compile_option("ghdl.flags"), ["--xyz"])

    def test_add_compile_option_does_not_mutate_argument(self):
        self.project.add_library("lib", "lib_path")
        file1 = self.add_source_file("lib", "file.vhd", "")
        options = ["--foo"]
        file1.add_compile_option("ghdl.flags", options)
        options[0] = "--xyz"
        self.assertEqual(file1.get_compile_option("ghdl.flags"), ["--foo"])
        file1.add_compile_option("ghdl.flags", ["--bar"])
        self.assertEqual(options, ["--xyz"])

    def test_set_compile_option_does_not_mutate_argument(self):
        self.project.add_library("lib", "lib_path")
        file1 = self.add_source_file("lib", "file.vhd", "")
        options = ["--foo"]
        file1.set_compile_option("ghdl.flags", options)
        options[0] = "--xyz"
        self.assertEqual(file1.get_compile_option("ghdl.flags"), ["--foo"])

    def test_compile_option_validation(self):
        self.project.add_library("lib", "lib_path")
        source_file = self.add_source_file("lib", "file.vhd", "")
        self.assertRaises(ValueError, source_file.set_compile_option, "foo", None)
        self.assertRaises(ValueError, source_file.set_compile_option, "ghdl.flags", None)
        self.assertRaises(ValueError, source_file.add_compile_option, "ghdl.flags", None)
        self.assertRaises(ValueError, source_file.get_compile_option, "foo")

    def test_should_recompile_files_affected_by_change_with_later_timestamp(self):
        file1, file2, file3 = self.create_dummy_three_file_project()

        self.update(file1)
        self.update(file2)
        self.update(file3)
        self.assert_should_recompile([])

        file1, file2, file3 = self.create_dummy_three_file_project()
        self.assert_should_recompile([])

        file1, file2, file3 = self.create_dummy_three_file_project(update_file1=True)
        self.assert_should_recompile([file1, file2, file3])

        tick()
        self.update(file1)
        self.assert_should_recompile([file2, file3])

    def test_should_recompile_files_missing_hash(self):
        file1, file2, file3 = self.create_dummy_three_file_project()

        self.update(file1)
        self.update(file2)
        self.update(file3)
        self.assert_should_recompile([])

        os.remove(self.hash_file_name_of(file2))
        self.assert_should_recompile([file2, file3])

    def test_finds_component_instantiation_dependencies(self):
        self.project.add_library("toplib", "work_path")
        top = self.add_source_file(
            "toplib",
            "top.vhd",
            """\
entity top is
end entity;

architecture arch of top is
begin
    labelFoo : component foo
    generic map(WIDTH => 16)
    port map(clk => '1',
             rst => '0',
             in_vec => record_reg.input_signal,
             output => some_signal(UPPER_CONSTANT-1 downto LOWER_CONSTANT+1));

    label2Foo : foo2
    port map(clk => '1',
             rst => '0',
             output => "00");
end architecture;
""",
        )

        comp1 = self.add_source_file(
            "toplib",
            "comp1.vhd",
            """\
entity foo is
end entity;
""",
        )

        comp2 = self.add_source_file(
            "toplib",
            "comp2.vhd",
            """\
entity foo2 is
end entity;

architecture arch of foo2 is
begin
end architecture;
""",
        )
        comp1_arch = self.add_source_file(
            "toplib",
            "comp1_arch.vhd",
            """\
architecture arch of foo is
begin
end architecture;
""",
        )

        self.assert_has_component_instantiation("top.vhd", "foo")
        self.assert_has_component_instantiation("top.vhd", "foo2")
        dependencies = self.project.get_dependencies_in_compile_order([top], implementation_dependencies=True)
        self.assertIn(comp1, dependencies)
        self.assertIn(comp1_arch, dependencies)
        self.assertIn(comp2, dependencies)

    def test_get_minimal_file_set_in_compile_order_without_target(self):
        self.create_dummy_three_file_project()
        deps = self.project.get_minimal_file_set_in_compile_order()
        self.assertEqual(len(deps), 3)
        self.assertTrue(deps[0] == self.project.get_source_files_in_order()[0])
        self.assertTrue(deps[1] == self.project.get_source_files_in_order()[1])
        self.assertTrue(deps[2] == self.project.get_source_files_in_order()[2])

    def test_get_minimal_file_set_in_compile_order_with_target(self):
        self.create_dummy_three_file_project()
        deps = self.project.get_minimal_file_set_in_compile_order(
            target_files=[self.project.get_source_files_in_order()[1]]
        )
        self.assertEqual(len(deps), 2)
        self.assertTrue(deps[0] == self.project.get_source_files_in_order()[0])
        self.assertTrue(deps[1] == self.project.get_source_files_in_order()[1])

        # To test that indirect dependencies are included
        deps = self.project.get_dependencies_in_compile_order(
            target_files=[self.project.get_source_files_in_order()[2]]
        )
        self.assertEqual(len(deps), 3)
        self.assertTrue(deps[0] == self.project.get_source_files_in_order()[0])
        self.assertTrue(deps[1] == self.project.get_source_files_in_order()[1])
        self.assertTrue(deps[2] == self.project.get_source_files_in_order()[2])

    def test_get_dependencies_in_compile_order_without_target(self):
        self.create_dummy_three_file_project()
        deps = self.project.get_dependencies_in_compile_order()
        self.assertEqual(len(deps), 3)
        self.assertTrue(deps[0] == self.project.get_source_files_in_order()[0])
        self.assertTrue(deps[1] == self.project.get_source_files_in_order()[1])
        self.assertTrue(deps[2] == self.project.get_source_files_in_order()[2])

    def test_get_dependencies_in_compile_order_with_target(self):
        self.create_dummy_three_file_project()
        deps = self.project.get_dependencies_in_compile_order(
            target_files=[self.project.get_source_files_in_order()[1]]
        )
        self.assertEqual(len(deps), 2)
        self.assertTrue(deps[0] == self.project.get_source_files_in_order()[0])
        self.assertTrue(deps[1] == self.project.get_source_files_in_order()[1])

        # To test that indirect dependencies are included
        deps = self.project.get_dependencies_in_compile_order(
            target_files=[self.project.get_source_files_in_order()[2]]
        )
        self.assertEqual(len(deps), 3)
        self.assertTrue(deps[0] == self.project.get_source_files_in_order()[0])
        self.assertTrue(deps[1] == self.project.get_source_files_in_order()[1])
        self.assertTrue(deps[2] == self.project.get_source_files_in_order()[2])

    def test_compiles_same_file_into_different_libraries(self):
        pkgs = []
        second_pkgs = []
        self.project.add_library("lib", "lib_path")

        other_pkg = self.add_source_file(
            "lib",
            "other_pkg.vhd",
            """
package other_pkg is
end package other_pkg;
""",
        )

        for lib in ["lib1", "lib2"]:
            self.project.add_library(lib, lib + "_path")
            pkgs.append(
                self.add_source_file(
                    lib,
                    "pkg.vhd",
                    """
library lib;
use lib.other_pkg.all;

package pkg is
end package pkg;
""",
                )
            )

            second_pkgs.append(
                self.add_source_file(
                    lib,
                    lib + "_pkg.vhd",
                    """
use work.pkg.all;

package second_pkg is
end package second_pkg;
""",
                )
            )

        self.assertNotEqual(self.hash_file_name_of(pkgs[0]), self.hash_file_name_of(pkgs[1]))
        self.assertEqual(len(self.project.get_files_in_compile_order()), 5)
        self.assert_compiles(other_pkg, before=pkgs[0])
        self.assert_compiles(other_pkg, before=pkgs[1])
        self.assert_compiles(pkgs[0], before=second_pkgs[0])
        self.assert_compiles(pkgs[1], before=second_pkgs[1])

    def test_has_verilog_module(self):
        self.project.add_library("lib", "lib_path")
        self.add_source_file(
            "lib",
            "module.v",
            """\
module name;
endmodule
""",
        )
        library = self.project.get_library("lib")
        modules = library.get_modules()
        self.assertEqual(len(modules), 1)

    def test_finds_verilog_package_import_dependencies(self):
        self.project.add_library("lib", "lib_path")
        pkg = self.add_source_file(
            "lib",
            "pkg.sv",
            """\
package pkg;
endpackage
""",
        )
        module = self.add_source_file(
            "lib",
            "module.sv",
            """\
module name;
  import pkg::*;
endmodule
""",
        )
        self.assert_compiles(pkg, before=module)

    def test_finds_verilog_package_reference_dependencies(self):
        self.project.add_library("lib", "lib_path")
        pkg = self.add_source_file(
            "lib",
            "pkg.sv",
            """\
package pkg;
endpackage
""",
        )
        module = self.add_source_file(
            "lib",
            "module.sv",
            """\
module name;
  pkg::func();
endmodule
""",
        )
        self.assert_compiles(pkg, before=module)

    def test_verilog_package_reference_is_case_sensitive(self):
        self.project = Project()
        self.project.add_library("lib", "lib_path")
        pkg = self.add_source_file(
            "lib",
            "pkg.sv",
            """\
package Pkg;
endpackage
""",
        )
        module = self.add_source_file(
            "lib",
            "module.sv",
            """\
module name;
  pkg::func();
endmodule
""",
        )
        self.assert_not_compiles(pkg, before=module)

        self.project = Project()
        self.project.add_library("lib", "lib_path")
        pkg = self.add_source_file(
            "lib",
            "pkg.sv",
            """\
package pkg;
endpackage
""",
        )
        module = self.add_source_file(
            "lib",
            "module.sv",
            """\
module name;
  Pkg::func();
endmodule
""",
        )
        self.assert_not_compiles(pkg, before=module)

    def test_finds_verilog_module_instantiation_dependencies(self):
        self.project.add_library("lib", "lib_path")
        module1 = self.add_source_file(
            "lib",
            "module1.sv",
            """\
module module1;
endmodule
""",
        )
        module2 = self.add_source_file(
            "lib",
            "module2.sv",
            """\
module module2;
  module1 inst();
endmodule
""",
        )
        self.assert_compiles(module1, before=module2)

    def test_verilog_module_instantiation_is_case_sensitive(self):
        self.project = Project()
        self.project.add_library("lib", "lib_path")
        module1 = self.add_source_file(
            "lib",
            "module1.sv",
            """\
module Module1;
endmodule
""",
        )
        module2 = self.add_source_file(
            "lib",
            "module2.sv",
            """\
module module2;
  module1 inst();
endmodule
""",
        )
        self.assert_not_compiles(module1, before=module2)

        self.project = Project()
        self.project.add_library("lib", "lib_path")
        module1 = self.add_source_file(
            "lib",
            "module1.sv",
            """\
module module1;
endmodule
""",
        )
        module2 = self.add_source_file(
            "lib",
            "module2.sv",
            """\
module module2;
  Module1 inst();
endmodule
""",
        )
        self.assert_not_compiles(module1, before=module2)

    def test_finds_verilog_module_instantiation_dependencies_in_vhdl(self):
        self.project.add_library("lib1", "lib_path")
        self.project.add_library("lib2", "lib_path")
        module1 = self.add_source_file(
            "lib1",
            "module1.sv",
            """\
module module1;
endmodule
""",
        )
        module2 = self.add_source_file(
            "lib2",
            "module2.vhd",
            """\
library lib1;

entity ent is
end entity;

architecture a of ent is
begin
  inst : entity lib1.module1;
end architecture;
""",
        )
        self.assert_compiles(module1, before=module2)

    def test_finds_verilog_include_dependencies(self):
        def create_project():
            """
            Create the test project
            """
            self.project = Project()
            self.project.add_library("lib", "lib_path")
            return self.add_source_file(
                "lib",
                "module.sv",
                """\
`include "include.svh"
""",
            )

        write_file(
            "include.svh",
            """\
module name;
endmodule
""",
        )
        module = create_project()
        self.assert_should_recompile([module])

        for src_file in self.project.get_files_in_compile_order():
            self.update(src_file)
        create_project()
        self.assert_should_recompile([])

        write_file(
            "include.svh",
            """\
module other_name;
endmodule
""",
        )
        module = create_project()
        self.assert_should_recompile([module])

    def test_verilog_defines_affects_dependency_scanning(self):
        self.project.add_library("lib", "lib_path")
        self.add_source_file(
            "lib",
            "module.v",
            """\
`ifdef foo
module mod;
endmodule
`endif
""",
        )
        library = self.project.get_library("lib")
        modules = library.get_modules()
        self.assertEqual(len(modules), 0)

        self.project = Project()
        self.project.add_library("lib", "lib_path")
        self.add_source_file(
            "lib",
            "module.v",
            """\
`ifdef foo
module mod;
endmodule
`endif
""",
            defines={"foo": ""},
        )
        library = self.project.get_library("lib")
        modules = library.get_modules()
        self.assertEqual(len(modules), 1)

    def test_recompile_when_updating_defines(self):
        contents1 = """
module mod1;
endmodule
"""
        contents2 = """
module mod2;
endmodule
"""
        self.project = Project()
        self.project.add_library("lib", "lib_path")
        mod1 = self.add_source_file("lib", "module1.v", contents1)
        mod2 = self.add_source_file("lib", "module2.v", contents2)
        self.assert_should_recompile([mod1, mod2])
        self.update(mod1)
        self.update(mod2)
        self.assert_should_recompile([])

        self.project = Project()
        self.project.add_library("lib", "lib_path")
        mod1 = self.add_source_file("lib", "module1.v", contents1, defines={"foo": "bar"})
        mod2 = self.add_source_file("lib", "module2.v", contents2)
        self.assert_should_recompile([mod1])
        self.update(mod1)
        self.update(mod2)
        self.assert_should_recompile([])

        self.project = Project()
        self.project.add_library("lib", "lib_path")
        mod1 = self.add_source_file("lib", "module1.v", contents1, defines={"foo": "other_bar"})
        mod2 = self.add_source_file("lib", "module2.v", contents2)
        self.assert_should_recompile([mod1])
        self.update(mod1)
        self.update(mod2)
        self.assert_should_recompile([])

    def test_manual_dependencies(self):
        self.project.add_library("lib", "lib_path")
        ent1 = self.add_source_file(
            "lib",
            "ent1.vhd",
            """\
entity ent1 is
end ent1;

architecture arch of ent1 is
begin
end architecture;
""",
        )

        ent2 = self.add_source_file(
            "lib",
            "ent2.vhd",
            """\
entity ent2 is
end ent2;

architecture arch of ent2 is
begin
end architecture;
""",
        )

        self.project.add_manual_dependency(ent2, depends_on=ent1)
        self.assert_compiles(ent1, before=ent2)

    @mock.patch("vunit.project.LOGGER", autospec=True)
    def test_circular_dependencies_causes_error(self, logger):
        self.project.add_library("lib", "lib_path")
        self.add_source_file(
            "lib",
            "ent1.vhd",
            """\
entity ent1 is
end ent1;

architecture arch of ent1 is
begin
   ent2_inst : entity work.ent2;
end architecture;
""",
        )

        self.add_source_file(
            "lib",
            "ent2.vhd",
            """\
entity ent2 is
end ent2;

architecture arch of ent2 is
begin
   ent1_inst : entity work.ent1;
end architecture;
""",
        )

        self.assertRaises(CompileError, self.project.get_files_in_compile_order)
        logger.error.assert_called_once_with(
            "Found circular dependency:\n%s", "ent1.vhd ->\n" "ent2.vhd ->\n" "ent1.vhd"
        )

    def test_order_of_adding_libraries_is_kept(self):
        for order in itertools.combinations(range(4), 4):
            project = Project()
            for idx in order:
                project.add_library("lib%i" % idx, "lib%i_path" % idx)

            library_names = [lib.name for lib in project.get_libraries()]
            self.assertEqual(library_names, ["lib%i" % idx for idx in order])

    def test_file_type_of(self):
        self.assertEqual(file_type_of("file.vhd"), "vhdl")
        self.assertEqual(file_type_of("file.vhdl"), "vhdl")
        self.assertEqual(file_type_of("file.sv"), "systemverilog")
        self.assertEqual(file_type_of("file.v"), "verilog")
        self.assertEqual(file_type_of("file.vams"), "verilog")
        self.assertRaises(RuntimeError, file_type_of, "file.foo")

    def test_circular_dependencies_through_libraries(self):
        """
        Create a projected containing two identical files in two separated
        library and instantiate an entity from the first library.
        """
        self.project = Project()
        self.project.add_library("lib_1", "work_path")
        self.project.add_library("lib_2", "work_path")
        self.project.add_library("lib", "work_path")
        text_file_1_2 = """\
        library ieee;
        use ieee.std_logic_1164.all;

        entity buffer1 is
          port (Q : out std_logic);
        end entity;

        architecture arch of buffer1 is begin
          Q <= '1';
        end architecture;

        library ieee;
        use ieee.std_logic_1164.all;

        entity buffer2 is
          port (Q : out std_logic);
        end entity;

        architecture arch of buffer2 is
          component buffer1
            port (Q : out std_logic);
          end component buffer1;

        begin
          my_buffer_i : buffer1
            port map (Q => Q);
        end architecture;
        """
        self.add_source_file("lib_1", "file1.vhd", text_file_1_2)
        self.add_source_file("lib_2", "file2.vhd", text_file_1_2)
        file3 = self.add_source_file(
            "lib",
            "file3.vhd",
            """\
        library lib_1;

        entity your_buffer is
        end entity;

        architecture arch of your_buffer is
        begin
          my_buffer_i : entity lib_1.buffer1;
        end architecture;
        """,
        )
        self.project.get_dependencies_in_compile_order([file3], implementation_dependencies=True)

    def test_dependencies_on_multiple_libraries(self):
        """
        Create a projected containing two identical files in two separated
        library and instantiate an entity from the first library.
        """
        self.project = Project()
        self.project.add_library("lib_1", "work_path")
        self.project.add_library("lib_2", "work_path")
        self.project.add_library("lib", "work_path")
        text_file_1_2 = """\
library ieee;use ieee.std_logic_1164.all;
entity buffer1 is  port (D : in std_logic;Q : out std_logic);end entity;
architecture arch of buffer1 is begin Q <= D; end architecture;
library ieee;use ieee.std_logic_1164.all;
entity buffer2 is port (D : in std_logic; Q : out std_logic);end entity;
architecture arch of buffer2 is
component buffer1 port (D : in  std_logic;Q : out std_logic);end component buffer1;
begin my_buffer_i : buffer1 port map (D => D,Q => Q);end architecture;
        """
        self.add_source_file("lib_1", "file1.vhd", text_file_1_2)
        lib2_file1_vhd = self.add_source_file("lib_2", "file1.vhd", text_file_1_2)
        file3 = self.add_source_file(
            "lib",
            "file3.vhd",
            """\
library ieee;use ieee.std_logic_1164.all;
library lib_1;entity your_buffer is port (D : in std_logic; Q : out std_logic);end entity;
architecture arch of your_buffer is
component buffer1 port (D : in  std_logic;Q : out std_logic);end component buffer1;
begin  my_buffer_i : buffer1 port map (D => D,Q => Q);end architecture;
        """,
        )
        dep_files = self.project.get_dependencies_in_compile_order([file3], implementation_dependencies=True)
        self.assertNotIn(lib2_file1_vhd, dep_files)

    def test_dependencies_on_separated_architecture(self):
        """
        Create a projected containing an entity file separated from architecture file.
        Dependency should involve also architecture.
        """
        self.project = Project()
        self.project.add_library("lib", "work_path")
        self.add_source_file(
            "lib",
            "file1.vhd",
            """\
library ieee;
use ieee.std_logic_1164.all;

entity buffer1 is
  port (D : in std_logic;
        Q : out std_logic);
end entity;
        """,
        )

        file1_arch_vhd = self.add_source_file(
            "lib",
            "file1_arch.vhd",
            """\
library ieee;
use ieee.std_logic_1164.all;

architecture arch of buffer1 is
begin
  Q <= D;
end architecture;
        """,
        )

        file3 = self.add_source_file(
            "lib",
            "file3.vhd",
            """\
library ieee;
use ieee.std_logic_1164.all;

entity your_buffer is
port (D : in std_logic; Q : out std_logic);
end entity;

architecture arch of your_buffer is
begin
my_buffer_i : entity work.buffer1
  port map (D => D,Q => Q);
end architecture;
        """,
        )
        dep_files = self.project.get_dependencies_in_compile_order([file3], implementation_dependencies=True)
        self.assertIn(file1_arch_vhd, dep_files)

    def test_dependencies_on_verilog_component(self):
        """
        Create a projected containing an verilog file separated.
        Dependency should involve it.
        """
        self.project = Project()
        self.project.add_library("lib", "work_path")
        file1_v = self.add_source_file(
            "lib",
            "file1.v",
            """\
module buffer1 (input   D,output   Q);
assign Q = D;
endmodule
        """,
        )
        file3 = self.add_source_file(
            "lib",
            "file3.vhd",
            """\
library ieee;use ieee.std_logic_1164.all;
entity your_buffer is port (D : in std_logic;Q : out std_logic);end entity;
architecture arch of your_buffer is
component buffer1 port (D : in  std_logic;Q : out std_logic);end component buffer1;
begin my_buffer_i : buffer1 port map (D => D,Q => Q);end architecture;
        """,
        )
        dep_files = self.project.get_dependencies_in_compile_order([file3], implementation_dependencies=True)
        self.assertIn(file1_v, dep_files)

    def create_dummy_three_file_project(self, update_file1=False):
        """
        Create a projected containing three dummy files
        optionally only updating file1
        """
        self.project = Project()
        self.project.add_library("lib", "work_path")

        if update_file1:
            file1 = self.add_source_file(
                "lib",
                "file1.vhd",
                """\
entity module1 is
end entity;

architecture arch of module1 is
begin
end architecture;
""",
            )
        else:
            file1 = self.add_source_file(
                "lib",
                "file1.vhd",
                """\
entity module1 is
end entity;

architecture arch of module1 is
begin
  report "Updated";
end architecture;
""",
            )
        file2 = self.add_source_file(
            "lib",
            "file2.vhd",
            """\
entity module2 is
end entity;

architecture arch of module2 is
begin
  module1_inst : entity lib.module1;
end architecture;
""",
        )

        file3 = self.add_source_file(
            "lib",
            "file3.vhd",
            """\
entity module3 is
end entity;

architecture arch of module3 is
begin
  module1_inst : entity work.module2;
end architecture;
""",
        )
        return file1, file2, file3

    def test_add_source_file_has_vhdl_standard(self):
        write_file("file.vhd", "")

        for std in ("93", "2002", "2008", "2019"):
            project = Project()
            project.add_library("lib", "lib_path")
            source_file = project.add_source_file("file.vhd", library_name="lib", file_type="vhdl", vhdl_standard=std)
            self.assertEqual(source_file.get_vhdl_standard(), std)

    def test_add_source_file_has_no_parse_vhdl(self):

        for no_parse in (True, False):
            project = Project()
            file_name = "file.vhd"
            write_file(
                file_name,
                """
    entity ent is
    end entity;
                       """,
            )
            project.add_library("lib", "work_path")
            source_file = project.add_source_file(
                file_name, "lib", file_type=file_type_of(file_name), no_parse=no_parse
            )
            self.assertEqual(len(source_file.design_units), int(not no_parse))

    def test_add_source_file_has_no_parse_verilog(self):

        for no_parse in (True, False):
            project = Project()
            file_name = "file.v"
            write_file(
                file_name,
                """
    module mod;
    endmodule
                       """,
            )
            project.add_library("lib", "work_path")
            source_file = project.add_source_file(
                file_name, "lib", file_type=file_type_of(file_name), no_parse=no_parse
            )
            self.assertEqual(len(source_file.design_units), int(not no_parse))

    @mock.patch("vunit.project.LOGGER")
    def test_no_warning_builtin_library_reference(self, mock_logger):
        self.project.add_library("lib", "lib_path")

        self.add_source_file(
            "lib",
            "ent.vhd",
            """
use std.foo.all;
use ieee.bar.all;
use builtin_lib.all;
""",
        )

        self.project.add_builtin_library("builtin_lib")
        self.project.get_files_in_compile_order()
        warning_calls = mock_logger.warning.call_args_list
        self.assertEqual(len(warning_calls), 0)

    def test_add_external_library(self):
        os.makedirs("lib_path")
        self.project.add_library("lib", "lib_path", is_external=True)

    def test_add_external_library_must_exist(self):
        try:
            self.project.add_library("lib2", "lib_path2", is_external=True)
        except ValueError as err:
            self.assertEqual(str(err), "External library 'lib_path2' does not exist")
        else:
            assert False, "ValueError not raised"

    def test_add_external_library_must_be_a_directory(self):
        write_file("lib_path3", "")
        try:
            self.project.add_library("lib3", "lib_path3", is_external=True)
        except ValueError as err:
            self.assertEqual(str(err), "External library must be a directory. Got 'lib_path3'")
        else:
            assert False, "ValueError not raised"

    def add_source_file(self, library_name, file_name, contents, defines=None):
        """
        Convenient wrapper arround project.add_source_file
        """
        write_file(file_name, contents)
        source_file = self.project.add_source_file(
            file_name, library_name, file_type=file_type_of(file_name), defines=defines
        )
        return source_file

    def hash_file_name_of(self, source_file):
        """
        Get the hash file name of a source_file
        """
        return self.project._hash_file_name_of(source_file)  # pylint: disable=protected-access

    def update(self, source_file):
        """
        Wrapper arround project.update
        """
        self.project.update(source_file)

    def assert_should_recompile(self, source_files):
        self.assertCountEqual(source_files, self.project.get_files_in_compile_order())

    def assert_compiles(self, source_file, before):
        """
        Assert that the compile order of source_file is before the file named 'before'.
        """
        for src_file in self.project.get_files_in_compile_order():
            self.update(src_file)
        self.assert_should_recompile([])
        tick()
        self.update(source_file)
        self.assertIn(before, self.project.get_files_in_compile_order())

    def assert_not_compiles(self, source_file, before):
        """
        Assert that the compile order of source_file is not before the file named 'before'.
        """
        for src_file in self.project.get_files_in_compile_order():
            self.update(src_file)
        self.assert_should_recompile([])
        tick()
        self.update(source_file)
        self.assertNotIn(before, self.project.get_files_in_compile_order())

    def assert_has_package_body(self, source_file_name, package_name):
        """
        Assert that there is a package body with package_name withing source_file_name
        """
        unit = self._find_design_unit(source_file_name, "package body", package_name, False, package_name)
        self.assertIsNotNone(unit)

    def assert_has_package(self, source_file_name, name):
        """
        Assert that there is a package with name withing source_file_name
        """
        unit = self._find_design_unit(source_file_name, "package", name)
        self.assertIsNotNone(unit)

    def assert_has_entity(self, source_file, name, generic_names=None, architecture_names=None):
        """
        Assert that there is an entity with name withing source_file
        that has architectures with architecture_names.
        """
        generic_names = [] if generic_names is None else generic_names
        architecture_names = [] if architecture_names is None else architecture_names

        for entity in source_file.library.get_entities():
            if entity.name == name:
                self.assertCountEqual(entity.generic_names, generic_names)
                self.assertCountEqual(entity.architecture_names, architecture_names)
                return

        self.assertFalse("Did not find entity " + name + "in " + source_file.name)

    def assert_has_architecture(self, source_file_name, name, entity_name):
        """
        Assert that there is an architecture with name of entity_name within source_file_name
        """
        unit = self._find_design_unit(source_file_name, "architecture", name, False, entity_name)
        self.assertIsNotNone(unit)

    def assert_has_component_instantiation(self, source_file_name, component_name):
        """
        Assert that there is a component instantion with component with source_file_name
        """
        found_comp = False
        for source_file in self.project.get_source_files_in_order():
            for component in source_file.depending_components:
                if component == component_name:
                    found_comp = True

        self.assertTrue(
            found_comp,
            "Did not find component " + component_name + " in " + source_file_name,
        )

    def _find_design_unit(  # pylint: disable=too-many-arguments
        self,
        source_file_name,
        design_unit_type,
        design_unit_name,
        is_primary=True,
        primary_design_unit_name=None,
    ):
        """
        Utility fnction to find and return a design unit
        """

        for source_file in self.project.get_source_files_in_order():
            for design_unit in source_file.design_units:
                if design_unit.unit_type != design_unit_type:
                    continue
                if design_unit.name != design_unit_name:
                    continue

                self.assertEqual(design_unit.is_primary, is_primary)
                self.assertEqual(source_file.name, source_file_name)
                if not is_primary:
                    self.assertEqual(design_unit.primary_design_unit, primary_design_unit_name)
                return design_unit

        return None


def tick():
    sleep(0.01)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

# pylint: disable=too-many-public-methods, too-many-lines

"""
Tests the test test_bench module
"""


import unittest
from pathlib import Path
from unittest import mock
from tests.common import with_tempdir, get_vhdl_test_bench
from vunit.test.bench import (
    TestBench,
    _remove_verilog_comments,
    _find_tests,
    _get_line_offsets,
    _lookup_lineno,
    _find_attributes,
    _find_tests_and_attributes,
    Test,
    FileLocation,
    Attribute,
    LegacyAttribute,
)
from vunit.configuration import AttributeException
from vunit.ostools import write_file


class TestTestBench(unittest.TestCase):
    """
    Tests the test bench
    """

    @with_tempdir
    def test_that_single_vhdl_test_is_created(self, tempdir):
        design_unit = Entity("tb_entity", file_name=str(Path(tempdir) / "file.vhd"))
        test_bench = TestBench(design_unit)
        tests = self.create_tests(test_bench)
        self.assert_has_tests(tests, ["lib.tb_entity.all"])

    @staticmethod
    @with_tempdir
    def test_no_architecture_at_creation(tempdir):
        design_unit = Entity("tb_entity", file_name=str(Path(tempdir) / "file.vhd"), no_arch=True)
        TestBench(design_unit)

    @with_tempdir
    def test_no_architecture_gives_runtime_error(self, tempdir):
        design_unit = Entity("tb_entity", file_name=str(Path(tempdir) / "file.vhd"), no_arch=True)
        test_bench = TestBench(design_unit)
        try:
            self.create_tests(test_bench)
        except RuntimeError as exc:
            self.assertEqual(str(exc), "Test bench 'tb_entity' has no architecture.")
        else:
            assert False, "RuntimeError not raised"

    @with_tempdir
    def test_that_single_verilog_test_is_created(self, tempdir):
        design_unit = Module("tb_module", file_name=str(Path(tempdir) / "file.v"))
        test_bench = TestBench(design_unit)
        tests = self.create_tests(test_bench)
        self.assert_has_tests(tests, ["lib.tb_module.all"])

    @with_tempdir
    def test_create_default_test(self, tempdir):
        design_unit = Entity("tb_entity", file_name=str(Path(tempdir) / "file.vhd"))
        design_unit.generic_names = ["runner_cfg"]
        test_bench = TestBench(design_unit)
        tests = self.create_tests(test_bench)
        self.assert_has_tests(tests, ["lib.tb_entity.all"])

    @with_tempdir
    def test_multiple_architectures_are_not_allowed_for_test_bench(self, tempdir):
        design_unit = Entity("tb_entity", file_name=str(Path(tempdir) / "file.vhd"))
        design_unit.generic_names = ["runner_cfg"]
        design_unit.add_architecture("arch2", file_name=str(Path(tempdir) / "arch2.vhd"))
        try:
            TestBench(design_unit)
        except RuntimeError as exc:
            self.assertEqual(
                str(exc),
                "Test bench not allowed to have multiple architectures. "
                "Entity tb_entity has arch:file.vhd, arch2:arch2.vhd",
            )
        else:
            assert False, "RuntimeError not raised"

    @with_tempdir
    def test_creates_tests_vhdl(self, tempdir):
        design_unit = Entity(
            "tb_entity",
            file_name=str(Path(tempdir) / "file.vhd"),
            contents="""\
if run("Test 1")
--if run("Test 2")
if run("Test 3")
if run("Test 4")
if run("Test 5") or run("Test 6")
if run  ("Test 7")
if run( "Test 8"  )
if ((run("Test 9")))
if my_protected_variable.run("Test 10")
""",
        )
        design_unit.generic_names = ["runner_cfg"]
        test_bench = TestBench(design_unit)
        tests = self.create_tests(test_bench)
        self.assert_has_tests(
            tests,
            [
                "lib.tb_entity.Test 1",
                "lib.tb_entity.Test 3",
                "lib.tb_entity.Test 4",
                "lib.tb_entity.Test 5",
                "lib.tb_entity.Test 6",
                "lib.tb_entity.Test 7",
                "lib.tb_entity.Test 8",
                "lib.tb_entity.Test 9",
            ],
        )

    @with_tempdir
    def test_creates_tests_verilog(self, tempdir):
        design_unit = Module(
            "tb_module",
            file_name=str(Path(tempdir) / "file.v"),
            contents="""\
`TEST_CASE("Test 1")
`TEST_CASE  ("Test 2")
`TEST_CASE( "Test 3"  )
// `TEST_CASE("Test 4")
/*
 `TEST_CASE("Test 5")
*/
/* `TEST_CASE("Test 6") */
`TEST_CASE("Test 7")
/* comment */
`TEST_CASE("Test 8")
/* comment */
""",
        )
        design_unit.generic_names = ["runner_cfg"]
        test_bench = TestBench(design_unit)
        tests = self.create_tests(test_bench)
        self.assert_has_tests(
            tests,
            [
                "lib.tb_module.Test 1",
                "lib.tb_module.Test 2",
                "lib.tb_module.Test 3",
                "lib.tb_module.Test 7",
                "lib.tb_module.Test 8",
            ],
        )

    @with_tempdir
    def test_keyerror_on_non_existent_test(self, tempdir):
        design_unit = Entity(
            "tb_entity",
            file_name=str(Path(tempdir) / "file.vhd"),
            contents='if run("Test")',
        )
        design_unit.generic_names = ["runner_cfg", "name"]
        test_bench = TestBench(design_unit)
        self.assertRaises(KeyError, test_bench.get_test_case, "Non existent test")

    @with_tempdir
    def test_creates_tests_when_adding_architecture_late(self, tempdir):
        design_unit = Entity("tb_entity", file_name=str(Path(tempdir) / "file.vhd"), no_arch=True)
        design_unit.generic_names = ["runner_cfg"]
        test_bench = TestBench(design_unit)

        design_unit.add_architecture(
            "arch",
            file_name=str(Path(tempdir) / "file.vhd"),
            contents="""\
if run("Test_1")
--if run("Test_2")
if run("Test_3")
""",
        )
        tests = self.create_tests(test_bench)
        self.assert_has_tests(tests, ["lib.tb_entity.Test_1", "lib.tb_entity.Test_3"])

    @with_tempdir
    def test_scan_tests_from_file(self, tempdir):
        design_unit = Entity("tb_entity", file_name=str(Path(tempdir) / "file.vhd"))
        design_unit.generic_names = ["runner_cfg"]
        test_bench = TestBench(design_unit)

        file_name = str(Path(tempdir) / "file.vhd")
        write_file(
            file_name,
            """\
if run("Test_1")
if run("Test_2")
""",
        )
        test_bench.scan_tests_from_file(file_name)
        tests = self.create_tests(test_bench)
        self.assert_has_tests(tests, ["lib.tb_entity.Test_1", "lib.tb_entity.Test_2"])

    def _test_scan_tests_from_file_location(self, tempdir, code):
        fstr = str(Path(tempdir) / "file.vhd")

        design_unit = Entity("tb_entity", file_name=fstr)
        design_unit.generic_names = ["runner_cfg"]
        test_bench = TestBench(design_unit)

        write_file(fstr, code)
        test_bench.scan_tests_from_file(fstr)
        tests = self.create_tests(test_bench)
        test_info = tests[0].test_information
        location = test_info["lib.tb_entity.Test_1"].location
        assert location.offset == code.find("Test_1")
        assert location.length == len("Test_1")

    @with_tempdir
    def test_scan_tests_from_file_location_unix(self, tempdir):
        self._test_scan_tests_from_file_location(tempdir, 'foo \n bar \n if run("Test_1")')

    @with_tempdir
    def test_scan_tests_from_file_location_dos(self, tempdir):
        self._test_scan_tests_from_file_location(tempdir, 'foo \r\n bar \r\n if run("Test_1")')

    @with_tempdir
    def test_scan_tests_from_missing_file(self, tempdir):
        design_unit = Entity("tb_entity", file_name=str(Path(tempdir) / "file.vhd"))
        design_unit.generic_names = ["runner_cfg"]
        test_bench = TestBench(design_unit)

        try:
            test_bench.scan_tests_from_file("missing.vhd")
        except ValueError as exc:
            self.assertEqual(str(exc), "File 'missing.vhd' does not exist")
        else:
            assert False, "ValueError not raised"

    @with_tempdir
    def test_does_not_add_all_suffix_with_named_configurations(self, tempdir):
        design_unit = Entity("tb_entity", file_name=str(Path(tempdir) / "file.vhd"))
        design_unit.generic_names = ["runner_cfg"]
        test_bench = TestBench(design_unit)

        tests = self.create_tests(test_bench)
        self.assert_has_tests(tests, ["lib.tb_entity.all"])

        test_bench.add_config(name="config", generics=dict())
        tests = self.create_tests(test_bench)
        self.assert_has_tests(tests, ["lib.tb_entity.config"])

    @with_tempdir
    def test_that_run_in_same_simulation_attribute_works(self, tempdir):
        design_unit = Entity(
            "tb_entity",
            file_name=str(Path(tempdir) / "file.vhd"),
            contents="""\
-- vunit: run_all_in_same_sim
if run("Test_1")
if run("Test_2")
--if run("Test_3")
""",
        )
        design_unit.generic_names = ["runner_cfg"]
        test_bench = TestBench(design_unit)
        tests = self.create_tests(test_bench)
        self.assert_has_tests(tests, [("lib.tb_entity", ("lib.tb_entity.Test_1", "lib.tb_entity.Test_2"))])

    @with_tempdir
    def test_add_config(self, tempdir):
        design_unit = Entity("tb_entity", file_name=str(Path(tempdir) / "file.vhd"))
        design_unit.generic_names = ["runner_cfg", "value", "global_value"]
        test_bench = TestBench(design_unit)

        test_bench.set_generic("global_value", "global value")

        test_bench.add_config(name="value=1", generics=dict(value=1, global_value="local value"))
        test_bench.add_config(name="value=2", generics=dict(value=2), attributes={".foo": "bar"})

        self.assertRaises(
            AttributeException,
            test_bench.add_config,
            name="c3",
            attributes={"foo", "bar"},
        )

        tests = self.create_tests(test_bench)
        self.assert_has_tests(tests, ["lib.tb_entity.value=1", "lib.tb_entity.value=2"])

        self.assertEqual(
            get_config_of(tests, "lib.tb_entity.value=1").generics,
            {"value": 1, "global_value": "local value"},
        )
        self.assertEqual(get_config_of(tests, "lib.tb_entity.value=1").attributes, {})
        self.assertEqual(
            get_config_of(tests, "lib.tb_entity.value=2").generics,
            {"value": 2, "global_value": "global value"},
        )
        self.assertEqual(get_config_of(tests, "lib.tb_entity.value=2").attributes, {".foo": "bar"})

    @with_tempdir
    def test_test_case_add_config(self, tempdir):
        design_unit = Entity(
            "tb_entity",
            file_name=str(Path(tempdir) / "file.vhd"),
            contents="""
if run("test 1")
if run("test 2")
        """,
        )
        design_unit.generic_names = ["runner_cfg", "value", "global_value"]
        test_bench = TestBench(design_unit)
        test_bench.set_generic("global_value", "global value")
        test_bench.set_sim_option("disable_ieee_warnings", True)

        test_case = test_bench.get_test_case("test 2")
        test_case.add_config(name="c1", generics=dict(value=1, global_value="local value"))
        test_case.add_config(
            name="c2",
            generics=dict(value=2),
            sim_options=dict(disable_ieee_warnings=False),
            attributes={".foo": "bar"},
        )

        self.assertRaises(
            AttributeException,
            test_case.add_config,
            name="c3",
            attributes={"foo", "bar"},
        )

        tests = self.create_tests(test_bench)
        self.assert_has_tests(
            tests,
            [
                "lib.tb_entity.test 1",
                "lib.tb_entity.c1.test 2",
                "lib.tb_entity.c2.test 2",
            ],
        )

        config_test1 = get_config_of(tests, "lib.tb_entity.test 1")
        config_c1_test2 = get_config_of(tests, "lib.tb_entity.c1.test 2")
        config_c2_test2 = get_config_of(tests, "lib.tb_entity.c2.test 2")
        self.assertEqual(config_test1.generics, {"global_value": "global value"})
        self.assertEqual(config_c1_test2.attributes, {})
        self.assertEqual(config_c1_test2.generics, {"value": 1, "global_value": "local value"})
        self.assertEqual(config_c2_test2.attributes, {".foo": "bar"})
        self.assertEqual(config_c2_test2.generics, {"value": 2, "global_value": "global value"})

    @with_tempdir
    def test_runtime_error_on_configuration_of_individual_test_with_same_sim(self, tempdir):
        design_unit = Entity(
            "tb_entity",
            file_name=str(Path(tempdir) / "file.vhd"),
            contents="""\
-- vunit: run_all_in_same_sim
if run("Test 1")
if run("Test 2")
""",
        )
        design_unit.generic_names = ["runner_cfg", "name"]
        test_bench = TestBench(design_unit)
        test_case = test_bench.get_test_case("Test 1")
        self.assertRaises(RuntimeError, test_case.set_generic, "name", "value")
        self.assertRaises(RuntimeError, test_case.set_sim_option, "name", "value")
        self.assertRaises(RuntimeError, test_case.add_config, "name")

    @with_tempdir
    def test_run_all_in_same_sim_can_be_configured(self, tempdir):
        design_unit = Entity(
            "tb_entity",
            file_name=str(Path(tempdir) / "file.vhd"),
            contents="""\
-- vunit: run_all_in_same_sim
if run("Test 1")
if run("Test 2")
""",
        )
        design_unit.generic_names = ["runner_cfg", "name"]
        test_bench = TestBench(design_unit)
        test_bench.set_generic("name", "value")
        test_bench.add_config("cfg")
        tests = self.create_tests(test_bench)
        self.assert_has_tests(
            tests,
            [
                (
                    "lib.tb_entity.cfg",
                    ("lib.tb_entity.cfg.Test 1", "lib.tb_entity.cfg.Test 2"),
                )
            ],
        )
        self.assertEqual(get_config_of(tests, "lib.tb_entity.cfg").generics, {"name": "value"})

    @with_tempdir
    def test_global_user_attributes_not_supported_yet(self, tempdir):
        design_unit = Entity(
            "tb_entity",
            file_name=str(Path(tempdir) / "file.vhd"),
            contents="""\
-- vunit: .attr0
if run("Test 1")
if run("Test 2")
""",
        )
        design_unit.generic_names = ["runner_cfg"]

        try:
            TestBench(design_unit)
        except RuntimeError as exc:
            self.assertEqual(
                str(exc),
                "File global attributes are not yet supported: .attr0 in %s line 1" % str(Path(tempdir) / "file.vhd"),
            )
        else:
            assert False, "RuntimeError not raised"

    @with_tempdir
    def test_error_on_global_attributes_on_tests(self, tempdir):
        design_unit = Entity(
            "tb_entity",
            file_name=str(Path(tempdir) / "file.vhd"),
            contents="""\
if run("Test 1")
-- vunit: run_all_in_same_sim
if run("Test 2")
""",
        )
        design_unit.generic_names = ["runner_cfg"]

        try:
            TestBench(design_unit)
        except RuntimeError as exc:
            self.assertEqual(
                str(exc),
                "Attribute run_all_in_same_sim is global and cannot be associated with test Test 1: %s line 2"
                % str(Path(tempdir) / "file.vhd"),
            )
        else:
            assert False, "RuntimeError not raised"

    @with_tempdir
    def test_test_information(self, tempdir):
        file_name = str(Path(tempdir) / "file.vhd")

        for same_sim in [True, False]:
            contents = get_vhdl_test_bench("tb_entity", tests=["Test 1", "Test 2"], same_sim=same_sim)

            design_unit = Entity("tb_entity", file_name=file_name, contents=contents)
            design_unit.generic_names = ["runner_cfg", "name"]
            test_bench = TestBench(design_unit)
            test_suites = self.create_tests(test_bench)

            if same_sim:
                self.assertEqual(len(test_suites), 1)
            else:
                self.assertEqual(len(test_suites), 2)

            self.assertEqual(
                set(item for test_suite in test_suites for item in test_suite.test_information.items()),
                set(
                    [
                        (
                            "lib.tb_entity.Test 1",
                            Test("Test 1", _file_location(file_name, "Test 1")),
                        ),
                        (
                            "lib.tb_entity.Test 2",
                            Test("Test 2", _file_location(file_name, "Test 2")),
                        ),
                    ]
                ),
            )

    @with_tempdir
    def test_fail_on_unknown_sim_option(self, tempdir):
        design_unit = Entity("tb_entity", file_name=str(Path(tempdir) / "file.vhd"))
        design_unit.generic_names = ["runner_cfg"]
        test_bench = TestBench(design_unit)
        self.assertRaises(ValueError, test_bench.set_sim_option, "unknown", "value")

    def test_remove_verilog_comments(self):
        self.assertEqual(_remove_verilog_comments("a\n// foo \nb"), "a\n       \nb")
        self.assertEqual(_remove_verilog_comments("a\n/* foo\n \n */ \nb"), "a\n      \n \n    \nb")

    def test_get_line_offsets(self):
        self.assertEqual(_get_line_offsets(""), [])
        self.assertEqual(_get_line_offsets("1"), [0])
        self.assertEqual(_get_line_offsets("12\n3"), [0, 3])
        self.assertEqual(_get_line_offsets("12\n3\n"), [0, 3])
        self.assertEqual(_get_line_offsets("12\n3\n4"), [0, 3, 5])

    def test_lookup_lineno(self):
        offsets = _get_line_offsets("12\n3\n4")
        self.assertEqual(_lookup_lineno(0, offsets), 1)
        self.assertEqual(_lookup_lineno(1, offsets), 1)
        self.assertEqual(_lookup_lineno(2, offsets), 1)
        self.assertEqual(_lookup_lineno(3, offsets), 2)
        self.assertEqual(_lookup_lineno(4, offsets), 2)
        self.assertEqual(_lookup_lineno(5, offsets), 3)

    def test_find_explicit_tests_vhdl(self):
        test1, test2 = _find_tests(
            """\
        -- if run("No test")
        if run("Test 1")
        if run("Test 2")
        """,
            file_name="file_name.vhd",
        )

        self.assertEqual(test1.name, "Test 1")
        self.assertEqual(test1.location.file_name, "file_name.vhd")
        self.assertEqual(test1.location.lineno, 2)

        self.assertEqual(test2.name, "Test 2")
        self.assertEqual(test2.location.file_name, "file_name.vhd")
        self.assertEqual(test2.location.lineno, 3)

    def test_find_explicit_tests_verilog(self):
        test1, test2 = _find_tests(
            """\
        /* `TEST_CASE("No test")

        */
        // `TEST_CASE("No test")
        `TEST_CASE("Test 1")
        `TEST_CASE("Test 2")
        """,
            file_name="file_name.sv",
        )

        self.assertEqual(test1.name, "Test 1")
        self.assertEqual(test1.location.file_name, "file_name.sv")
        self.assertEqual(test1.location.lineno, 5)

        self.assertEqual(test2.name, "Test 2")
        self.assertEqual(test2.location.file_name, "file_name.sv")
        self.assertEqual(test2.location.lineno, 6)

    def test_find_implicit_test_vhdl(self):
        with mock.patch("vunit.test.bench.LOGGER") as logger:
            (test,) = _find_tests(
                """\

            test_runner_setup(
            """,
                file_name="file_name.vhd",
            )
            self.assertEqual(test.name, None)
            self.assertEqual(test.location.file_name, "file_name.vhd")
            self.assertEqual(test.location.lineno, 2)
            assert not logger.warning.called

        with mock.patch("vunit.test.bench.LOGGER") as logger:
            (test,) = _find_tests(
                """\

            test_runner_setup
            """,
                file_name="file_name.vhd",
            )
            self.assertEqual(test.name, None)
            self.assertEqual(test.location.file_name, "file_name.vhd")
            self.assertEqual(test.location.lineno, 1)
            assert logger.warning.called

        with mock.patch("vunit.test.bench.LOGGER") as logger:
            (test,) = _find_tests(
                """\

            """,
                file_name="file_name.vhd",
            )
            self.assertEqual(test.name, None)
            self.assertEqual(test.location.file_name, "file_name.vhd")
            self.assertEqual(test.location.lineno, 1)
            assert logger.warning.called

    def test_find_implicit_test_verilog(self):
        with mock.patch("vunit.test.bench.LOGGER") as logger:
            (test,) = _find_tests(
                """\

            `TEST_SUITE""",
                file_name="file_name.sv",
            )
            self.assertEqual(test.name, None)
            self.assertEqual(test.location.file_name, "file_name.sv")
            self.assertEqual(test.location.lineno, 2)
            assert not logger.warning.called

        with mock.patch("vunit.test.bench.LOGGER") as logger:
            (test,) = _find_tests(
                """\
            TEST_SUITE
            """,
                file_name="file_name.sv",
            )
            self.assertEqual(test.name, None)
            self.assertEqual(test.location.file_name, "file_name.sv")
            self.assertEqual(test.location.lineno, 1)
            assert logger.warning.called

        with mock.patch("vunit.test.bench.LOGGER") as logger:
            (test,) = _find_tests(
                """\
            """,
                file_name="file_name.sv",
            )
            self.assertEqual(test.name, None)
            self.assertEqual(test.location.file_name, "file_name.sv")
            self.assertEqual(test.location.lineno, 1)
            assert logger.warning.called

    @mock.patch("vunit.test.bench.LOGGER")
    def test_duplicate_tests_cause_error(self, mock_logger):
        file_name = "file.vhd"
        self.assertRaises(
            RuntimeError,
            _find_tests,
            """\
if run("Test_1")
--if run("Test_1")
if run("Test_3")
if run("Test_2")
if run("Test_3")
if run("Test_3")
if run("Test_2")
        """,
            file_name=file_name,
        )

        error_calls = mock_logger.error.call_args_list
        self.assertEqual(len(error_calls), 3)

        msg = error_calls[0][0][0] % error_calls[0][0][1:]
        self.assertEqual(
            msg,
            'Duplicate test "Test_3" in %s line 5 previously defined on line 3' % file_name,
        )

        msg = error_calls[1][0][0] % error_calls[1][0][1:]
        self.assertEqual(
            msg,
            'Duplicate test "Test_3" in %s line 6 previously defined on line 3' % file_name,
        )

        msg = error_calls[2][0][0] % error_calls[2][0][1:]
        self.assertEqual(
            msg,
            'Duplicate test "Test_2" in %s line 7 previously defined on line 4' % file_name,
        )

    def test_find_attributes(self):
        code = """
// vunit: run_all_in_same_sim
// vunit:fail_on_warning
        """

        attributes = _find_attributes(code, file_name="file.vhd")
        self.assertEqual(
            attributes,
            [
                Attribute(
                    "run_all_in_same_sim",
                    None,
                    _code_file_location(code, "run_all_in_same_sim", "file.vhd"),
                ),
                Attribute(
                    "fail_on_warning",
                    None,
                    _code_file_location(code, "fail_on_warning", "file.vhd"),
                ),
            ],
        )

    def test_find_user_attributes(self):
        code = """
// vunit: .foo
// vunit: .foo-bar
        """
        attributes = _find_attributes(code, file_name="file.vhd")

        self.assertEqual(
            attributes,
            [
                Attribute(".foo", None, _code_file_location(code, ".foo", "file.vhd")),
                Attribute(".foo-bar", None, _code_file_location(code, ".foo-bar", "file.vhd")),
            ],
        )

    def test_invalid_attributes(self):
        try:
            _find_attributes(
                """\


// vunit: invalid
            """,
                file_name="file.vhd",
            )
        except RuntimeError as exc:
            self.assertEqual(str(exc), "Invalid attribute 'invalid' in file.vhd line 3")
        else:
            assert False, "RuntimeError not raised"

    def test_find_legacy_pragma(self):
        code = """
// vunit_pragma run_all_in_same_sim
// vunit_pragma fail_on_warning
        """
        attributes = _find_attributes(code, file_name="file.vhd")

        self.assertEqual(
            attributes,
            [
                LegacyAttribute(
                    "run_all_in_same_sim",
                    None,
                    _code_file_location(code, "run_all_in_same_sim", "file.vhd"),
                ),
                LegacyAttribute(
                    "fail_on_warning",
                    None,
                    _code_file_location(code, "fail_on_warning", "file.vhd"),
                ),
            ],
        )

    def test_associate_tests_and_attributes(self):
        (test1, test2), attributes = _find_tests_and_attributes(
            """\
// vunit: .arg0
        if run("test1")
// vunit: .arg1
// vunit: .arg1b
        if run("test2") // vunit: .arg2
// vunit: .arg2b
        """,
            file_name="file.vhd",
        )

        self.assertEqual([attr.name for attr in attributes], [".arg0"])

        self.assertEqual(test1.name, "test1")
        self.assertEqual(test1.location.file_name, "file.vhd")
        self.assertEqual(test1.location.lineno, 2)
        self.assertEqual([attr.name for attr in test1.attributes], [".arg1", ".arg1b"])

        self.assertEqual(test2.name, "test2")
        self.assertEqual(test2.location.file_name, "file.vhd")
        self.assertEqual(test2.location.lineno, 5)
        self.assertEqual([attr.name for attr in test2.attributes], [".arg2", ".arg2b"])

    def test_duplicate_attributes_ok(self):
        (test1, test2), attributes = _find_tests_and_attributes(
            """\
// vunit: .arg0
        if run("test1")
// vunit: .arg0
        if run("test2")
// vunit: .arg0
        """,
            file_name="file.vhd",
        )

        self.assertEqual([attr.name for attr in attributes], [".arg0"])

        self.assertEqual(test1.name, "test1")
        self.assertEqual(test1.location.file_name, "file.vhd")
        self.assertEqual(test1.location.lineno, 2)
        self.assertEqual([attr.name for attr in test1.attributes], [".arg0"])

        self.assertEqual(test2.name, "test2")
        self.assertEqual(test2.location.file_name, "file.vhd")
        self.assertEqual(test2.location.lineno, 4)
        self.assertEqual([attr.name for attr in test2.attributes], [".arg0"])

    def test_duplicate_test_attributes_not_ok(self):
        try:
            _find_tests_and_attributes(
                """\
        if run("test1")
// vunit: .arg0
// vunit: .arg0
        """,
                file_name="file.vhd",
            )
        except RuntimeError as exc:
            self.assertEqual(
                str(exc),
                "Duplicate attribute .arg0 of test test1 in file.vhd line 3, previously defined on line 2",
            )
        else:
            assert False, "RuntimeError not raised"

    def test_duplicate_global_attributes_not_ok(self):
        try:
            _find_tests_and_attributes(
                """\
// vunit: .arg0
// vunit: .arg0
        if run("test1")
        """,
                file_name="file.vhd",
            )
        except RuntimeError as exc:
            self.assertEqual(
                str(exc),
                "Duplicate attribute .arg0 of file.vhd line 2, previously defined on line 1",
            )
        else:
            assert False, "RuntimeError not raised"

    def test_does_not_associate_tests_and_legacy_attributes(self):
        code = """\
        if run("test1")
// vunit_pragma run_all_in_same_sim
// vunit_pragma fail_on_warning
        """
        (test1,), attributes = _find_tests_and_attributes(code, file_name="file.vhd")

        self.assertEqual(
            attributes,
            [
                LegacyAttribute(
                    "run_all_in_same_sim",
                    None,
                    _code_file_location(code, "run_all_in_same_sim", "file.vhd"),
                ),
                LegacyAttribute(
                    "fail_on_warning",
                    None,
                    _code_file_location(code, "fail_on_warning", "file.vhd"),
                ),
            ],
        )

        self.assertEqual(test1.name, "test1")
        self.assertEqual(test1.location.file_name, "file.vhd")
        self.assertEqual(test1.location.lineno, 1)
        self.assertEqual(test1.attributes, [])

    def assert_has_tests(self, test_list, tests):
        """
        Asser that the test_list contains tests.
        A test can be either a string to represent a single test or a
        tuple to represent multiple tests within a test suite.
        """
        self.assertEqual(len(test_list), len(tests))
        for test1, test2 in zip(test_list, tests):
            if isinstance(test2, tuple):
                name, test_cases = test2
                self.assertEqual(test1.name, name)
                self.assertEqual(test1.test_names, list(test_cases))
            else:
                self.assertEqual(test1.name, test2)
                self.assertEqual(test1.test_names, [test2])

    @staticmethod
    def create_tests(test_bench):
        """
        Helper method to reduce boiler plate
        """
        return test_bench.create_tests("simulator_if", elaborate_only=False)


def get_config_of(tests, test_name):
    """
    Find generic values of test
    """
    for test in tests:
        if test.name == test_name:
            try:
                return test._test_case._run._config  # pylint: disable=protected-access
            except AttributeError:
                return test._run._config  # pylint: disable=protected-access
    raise KeyError(test_name)


class Module(object):
    """
    Mock Module
    """

    def __init__(self, name, file_name, contents=""):
        self.name = name
        self.library_name = "lib"
        self.is_entity = False
        self.is_module = True
        self.generic_names = []
        self.file_name = file_name
        self.original_file_name = file_name
        write_file(file_name, contents)


class Entity(object):  # pylint: disable=too-many-instance-attributes
    """
    Mock Entity
    """

    def __init__(self, name, file_name, contents="", no_arch=False):
        self.name = name
        self.library_name = "lib"
        self.is_entity = True
        self.is_module = False
        self.architecture_names = {}

        if not no_arch:
            self.architecture_names = {"arch": file_name}

        self.generic_names = []
        self.file_name = file_name
        self.original_file_name = file_name
        write_file(file_name, contents)
        self._add_architecture_callback = None

    def set_add_architecture_callback(self, callback):
        """
        Set callback to be called when an architecture is added
        """
        assert self._add_architecture_callback is None
        self._add_architecture_callback = callback

    def add_architecture(self, name, file_name, contents=""):
        """
        Add architecture to this entity
        """
        self.architecture_names[name] = file_name
        write_file(file_name, contents)

        if self._add_architecture_callback is not None:
            self._add_architecture_callback()


def _code_file_location(code, string, file_name=None):
    """
    Create a FileLocation by finding the string within the code
    """
    offset = code.find(string)
    return FileLocation(
        file_name=file_name,
        offset=offset,
        length=len(string),
        lineno=1 + sum((1 for char in code[:offset] if char == "\n")),
    )


def _file_location(file_name, string):
    """
    Create a FileLocation by finding the string within file_name
    """

    with open(file_name, "r") as fptr:
        code = fptr.read()
        return _code_file_location(code, string, file_name)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
Interface towards Mentor Graphics ModelSim
"""

from pathlib import Path
import os
import logging
from configparser import RawConfigParser
from ..exceptions import CompileError
from ..ostools import Process, file_exists
from ..vhdl_standard import VHDL
from . import SimulatorInterface, ListOfStringOption, StringOption
from .vsim_simulator_mixin import VsimSimulatorMixin, fix_path

LOGGER = logging.getLogger(__name__)


class ModelSimInterface(VsimSimulatorMixin, SimulatorInterface):  # pylint: disable=too-many-instance-attributes
    """
    Mentor Graphics ModelSim interface

    The interface supports both running each simulation in separate vsim processes or
    re-using the same vsim process to avoid startup-overhead (persistent=True)
    """

    name = "modelsim"
    supports_gui_flag = True
    package_users_depend_on_bodies = False

    compile_options = [
        ListOfStringOption("modelsim.vcom_flags"),
        ListOfStringOption("modelsim.vlog_flags"),
    ]

    sim_options = [
        ListOfStringOption("modelsim.vsim_flags"),
        ListOfStringOption("modelsim.vsim_flags.gui"),
        ListOfStringOption("modelsim.init_files.after_load"),
        ListOfStringOption("modelsim.init_files.before_run"),
        StringOption("modelsim.init_file.gui"),
    ]

    @classmethod
    def from_args(cls, args, output_path, **kwargs):
        """
        Create new instance from command line arguments object
        """
        persistent = not (args.unique_sim or args.gui)

        return cls(
            prefix=cls.find_prefix(),
            output_path=output_path,
            persistent=persistent,
            gui=args.gui,
        )

    @classmethod
    def find_prefix_from_path(cls):
        """
        Find first valid modelsim toolchain prefix
        """

        def has_modelsim_ini(path):
            return os.path.isfile(str(Path(path).parent / "modelsim.ini"))

        return cls.find_toolchain(["vsim"], constraints=[has_modelsim_ini])

    @classmethod
    def supports_vhdl_package_generics(cls):
        """
        Returns True when this simulator supports VHDL package generics
        """
        return True

    @staticmethod
    def supports_coverage():
        """
        Returns True when the simulator supports coverage
        """
        return True

    def __init__(self, prefix, output_path, persistent=False, gui=False):
        SimulatorInterface.__init__(self, output_path, gui)
        VsimSimulatorMixin.__init__(
            self,
            prefix,
            persistent,
            sim_cfg_file_name=str(Path(output_path) / "modelsim.ini"),
        )
        self._libraries = []
        self._coverage_files = set()
        assert not (persistent and gui)
        self._create_modelsim_ini()

    def _create_modelsim_ini(self):
        """
        Create the modelsim.ini file
        """
        parent = str(Path(self._sim_cfg_file_name).parent)
        if not file_exists(parent):
            os.makedirs(parent)

        original_modelsim_ini = os.environ.get("VUNIT_MODELSIM_INI", str(Path(self._prefix).parent / "modelsim.ini"))
        with Path(original_modelsim_ini).open("rb") as fread:
            with Path(self._sim_cfg_file_name).open("wb") as fwrite:
                fwrite.write(fread.read())

    def add_simulator_specific(self, project):
        """
        Add libraries from modelsim.ini file and add coverage flags
        """
        mapped_libraries = self._get_mapped_libraries()
        for library_name in mapped_libraries:
            if not project.has_library(library_name):
                project.add_builtin_library(library_name)

    def setup_library_mapping(self, project):
        """
        Setup library mapping
        """
        mapped_libraries = self._get_mapped_libraries()

        for library in project.get_libraries():
            self._libraries.append(library)
            self.create_library(library.name, library.directory, mapped_libraries)

    def compile_source_file_command(self, source_file):
        """
        Returns the command to compile a single source file
        """
        if source_file.is_vhdl:
            return self.compile_vhdl_file_command(source_file)

        if source_file.is_any_verilog:
            return self.compile_verilog_file_command(source_file)

        LOGGER.error("Unknown file type: %s", source_file.file_type)
        raise CompileError

    @staticmethod
    def _std_str(vhdl_standard):
        """
        Convert standard to format of Modelsim command line flag
        """
        if vhdl_standard <= VHDL.STD_2008:
            return f"-{vhdl_standard!s}"

        raise ValueError(f"Invalid VHDL standard {vhdl_standard!s}")

    def compile_vhdl_file_command(self, source_file):
        """
        Returns the command to compile a vhdl file
        """
        return (
            [
                str(Path(self._prefix) / "vcom"),
                "-quiet",
                "-modelsimini",
                self._sim_cfg_file_name,
            ]
            + source_file.compile_options.get("modelsim.vcom_flags", [])
            + [
                self._std_str(source_file.get_vhdl_standard()),
                "-work",
                source_file.library.name,
                source_file.name,
            ]
        )

    def compile_verilog_file_command(self, source_file):
        """
        Returns the command to compile a verilog file
        """
        args = [
            str(Path(self._prefix) / "vlog"),
            "-quiet",
            "-modelsimini",
            self._sim_cfg_file_name,
        ]
        if source_file.is_system_verilog:
            args += ["-sv"]
        args += source_file.compile_options.get("modelsim.vlog_flags", [])
        args += ["-work", source_file.library.name, source_file.name]

        for library in self._libraries:
            args += ["-L", library.name]
        for include_dir in source_file.include_dirs:
            args += [f"+incdir+{include_dir!s}"]
        for key, value in source_file.defines.items():
            args += [f"+define+{key!s}={value!s}"]
        return args

    def create_library(self, library_name, path, mapped_libraries=None):
        """
        Create and map a library_name to path
        """
        mapped_libraries = mapped_libraries if mapped_libraries is not None else {}

        apath = str(Path(path).parent.resolve())

        if not file_exists(apath):
            os.makedirs(apath)

        if not file_exists(path):
            proc = Process([str(Path(self._prefix) / "vlib"), "-unix", path], env=self.get_env())
            proc.consume_output(callback=None)

        if library_name in mapped_libraries and mapped_libraries[library_name] == path:
            return

        cfg = parse_modelsimini(self._sim_cfg_file_name)
        cfg.set("Library", library_name, path)
        write_modelsimini(cfg, self._sim_cfg_file_name)

    def _get_mapped_libraries(self):
        """
        Get mapped libraries from modelsim.ini file
        """
        cfg = parse_modelsimini(self._sim_cfg_file_name)
        libraries = dict(cfg.items("Library"))
        if "others" in libraries:
            del libraries["others"]
        return libraries

    def _create_load_function(self, test_suite_name, config, output_path):
        """
        Create the vunit_load TCL function that runs the vsim command and loads the design
        """

        set_generic_str = " ".join(
            (
                f"-g/{config.entity_name!s}/{name!s}={encode_generic_value(value)!s}"
                for name, value in config.generics.items()
            )
        )
        pli_str = " ".join(f"-pli {{{fix_path(name)!s}}}" for name in config.sim_options.get("pli", []))

        if config.architecture_name is None:
            architecture_suffix = ""
        else:
            architecture_suffix = f"({config.architecture_name!s})"

        if config.sim_options.get("enable_coverage", False):
            coverage_file = str(Path(output_path) / "coverage.ucdb")
            self._coverage_files.add(coverage_file)
            coverage_save_cmd = (
                f"coverage save -onexit -testname {{{test_suite_name!s}}} -assert -directive "
                f"-cvg -codeAll {{{fix_path(coverage_file)!s}}}"
            )
            coverage_args = "-coverage"
        else:
            coverage_save_cmd = ""
            coverage_args = ""

        vsim_flags = [
            f"-wlf {{{fix_path(str(Path(output_path) / 'vsim.wlf'))!s}}}",
            "-quiet",
            "-t ps",
            # for correct handling of verilog fatal/finish
            "-onfinish stop",
            pli_str,
            set_generic_str,
            config.library_name + "." + config.entity_name + architecture_suffix,
            coverage_args,
            self._vsim_extra_args(config),
        ]

        # There is a known bug in modelsim that prevents the -modelsimini flag from accepting
        # a space in the path even with escaping, see issue #36
        if " " not in self._sim_cfg_file_name:
            vsim_flags.insert(0, f"-modelsimini {fix_path(self._sim_cfg_file_name)!s}")

        for library in self._libraries:
            vsim_flags += ["-L", library.name]

        vhdl_assert_stop_level_mapping = dict(warning=1, error=2, failure=3)

        tcl = """
proc vunit_load {{{{vsim_extra_args ""}}}} {{
    set vsim_failed [catch {{
        eval vsim ${{vsim_extra_args}} {{{vsim_flags}}}
    }}]

    if {{${{vsim_failed}}}} {{
       echo Command 'vsim ${{vsim_extra_args}} {vsim_flags}' failed
       echo Bad flag from vsim_extra_args?
       return true
    }}

    if {{[_vunit_source_init_files_after_load]}} {{
        return true
    }}

    global BreakOnAssertion
    set BreakOnAssertion {break_on_assert}

    global NumericStdNoWarnings
    set NumericStdNoWarnings {no_warnings}

    global StdArithNoWarnings
    set StdArithNoWarnings {no_warnings}

    {coverage_save_cmd}
    return false
}}
""".format(
            coverage_save_cmd=coverage_save_cmd,
            vsim_flags=" ".join(vsim_flags),
            break_on_assert=vhdl_assert_stop_level_mapping[config.vhdl_assert_stop_level],
            no_warnings=1 if config.sim_options.get("disable_ieee_warnings", False) else 0,
        )

        return tcl

    @staticmethod
    def _create_run_function():
        """
        Create the vunit_run function to run the test bench
        """
        return """

proc _vunit_run_failure {} {
    catch {
        # tb command can fail when error comes from pli
        echo "Stack trace result from 'tb' command"
        echo [tb]
        echo
        echo "Surrounding code from 'see' command"
        echo [see]
    }
}

proc _vunit_run {} {
    if {[_vunit_source_init_files_before_run]} {
        return true
    }

    proc on_break {} {
        resume
    }
    onbreak {on_break}

    run -all
}

proc _vunit_sim_restart {} {
    restart -f
}
"""

    def _vsim_extra_args(self, config):
        """
        Determine vsim_extra_args
        """
        vsim_extra_args = []
        vsim_extra_args = config.sim_options.get("modelsim.vsim_flags", vsim_extra_args)

        if self._gui:
            vsim_extra_args = config.sim_options.get("modelsim.vsim_flags.gui", vsim_extra_args)

        return " ".join(vsim_extra_args)

    def merge_coverage(self, file_name, args=None):
        """
        Merge coverage from all test cases
        """
        if self._persistent_shell is not None:
            # Teardown to ensure ucdb file was written.
            self._persistent_shell.teardown()

        if args is None:
            args = []

        coverage_files = str(Path(self._output_path) / "coverage_files.txt")
        vcover_cmd = [str(Path(self._prefix) / "vcover"), "merge", "-inputs"] + [coverage_files] + args + [file_name]
        with Path(coverage_files).open("w", encoding="utf-8") as fptr:
            for coverage_file in self._coverage_files:
                if file_exists(coverage_file):
                    fptr.write(str(coverage_file) + "\n")
                else:
                    LOGGER.warning("Missing coverage file: %s", coverage_file)

        print(f"Merging coverage files into {file_name!s}...")
        vcover_merge_process = Process(vcover_cmd, env=self.get_env())
        vcover_merge_process.consume_output()
        print("Done merging coverage files")

    @staticmethod
    def get_env():
        """
        Remove MODELSIM environment variable
        """
        remove = ("MODELSIM",)
        env = os.environ.copy()
        for key in remove:
            if key in env.keys():
                del env[key]
        return env


def encode_generic_value(value):
    """
    Ensure values with space in them are quoted
    """
    s_value = str(value)
    if " " in s_value:
        return f'{{"{s_value!s}"}}'
    if "," in s_value:
        return f'"{s_value!s}"'
    return s_value


def parse_modelsimini(file_name):
    """
    Parse a modelsim.ini file
    :returns: A RawConfigParser object
    """
    cfg = RawConfigParser()
    with Path(file_name).open("r", encoding="utf-8") as fptr:
        cfg.read_file(fptr)
    return cfg


def write_modelsimini(cfg, file_name):
    """
    Writes a modelsim.ini file
    """
    with Path(file_name).open("w", encoding="utf-8") as optr:
        cfg.write(optr)



# Classes to model a HDL design hierarchy
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
Functionality to represent and operate on a HDL code library
"""

import logging
from vunit.vhdl_standard import VHDLStandard

LOGGER = logging.getLogger(__name__)


class Library(object):  # pylint: disable=too-many-instance-attributes
    """
    Represents a VHDL library
    """

    def __init__(self, name: str, directory: str, vhdl_standard: VHDLStandard, is_external=False):
        self.name = name
        self.directory = directory

        # Default VHDL standard for files added unless explicitly set per file
        self.vhdl_standard = vhdl_standard

        self._source_files = {}  # type: ignore

        # Entity objects
        self._entities = {}  # type: ignore
        self._package_bodies = {}  # type: ignore

        self.primary_design_units = {}  # type: ignore

        # Entity name to architecture design unit mapping
        self._architectures = {}  # type: ignore

        # Verilog specific
        # Module objects
        self.modules = {}  # type: ignore
        self.verilog_packages = {}  # type: ignore

        self._is_external = is_external

    def add_source_file(self, source_file):
        """
        Add source file to library unless it exists

        returns The source file that has added or the old source file
        """
        if source_file.name in self._source_files:
            old_source_file = self._source_files[source_file.name]
            if old_source_file.content_hash != source_file.content_hash:
                raise RuntimeError(f"{source_file.name!s} already added to library {self.name!s}")

            LOGGER.info(
                "Ignoring duplicate file %s added to library %s due to identical contents",
                source_file.name,
                self.name,
            )

            return old_source_file

        self._source_files[source_file.name] = source_file
        source_file.add_to_library(self)

        return source_file

    def get_source_file(self, file_name):
        """
        Get source file with file name or raise KeyError
        """
        return self._source_files[file_name]

    @property
    def is_external(self):
        """
        External black box library, typically compiled outside of VUnit
        """
        return self._is_external

    @staticmethod
    def _warning_on_duplication(design_unit, old_file_name):
        """
        Utility function to give warning for design unit duplication
        """
        LOGGER.warning(
            "%s: %s '%s' previously defined in %s",
            design_unit.source_file.name,
            design_unit.unit_type,
            design_unit.name,
            old_file_name,
        )

    def _check_duplication(self, dictionary, design_unit):
        """
        Utility function to check if design_unit already in dictionary
        and give warning
        """
        if design_unit.name in dictionary:
            self._warning_on_duplication(design_unit, dictionary[design_unit.name].source_file.name)

    def add_vhdl_design_units(self, design_units):
        """
        Add VHDL design units to the library
        """
        for design_unit in design_units:
            if design_unit.is_primary:
                self._check_duplication(self.primary_design_units, design_unit)
                self.primary_design_units[design_unit.name] = design_unit

                if design_unit.unit_type == "entity":
                    if design_unit.name not in self._architectures:
                        self._architectures[design_unit.name] = {}
                    self._entities[design_unit.name] = design_unit

                    for architecture in self._architectures[design_unit.name].values():
                        design_unit.add_architecture(architecture)

            else:
                if design_unit.unit_type == "architecture":
                    if design_unit.primary_design_unit not in self._architectures:
                        self._architectures[design_unit.primary_design_unit] = {}

                    if design_unit.name in self._architectures[design_unit.primary_design_unit]:
                        self._warning_on_duplication(
                            design_unit,
                            self._architectures[design_unit.primary_design_unit][design_unit.name].source_file.name,
                        )

                    self._architectures[design_unit.primary_design_unit][design_unit.name] = design_unit

                    if design_unit.primary_design_unit in self._entities:
                        self._entities[design_unit.primary_design_unit].add_architecture(design_unit)

                if design_unit.unit_type == "package body":
                    if design_unit.primary_design_unit in self._package_bodies:
                        self._warning_on_duplication(
                            design_unit,
                            self._package_bodies[design_unit.primary_design_unit].source_file.name,
                        )
                    self._package_bodies[design_unit.primary_design_unit] = design_unit

    def add_verilog_design_units(self, design_units):
        """
        Add Verilog design units to the library
        """
        for design_unit in design_units:
            if design_unit.unit_type == "module":
                if design_unit.name in self.modules:
                    self._warning_on_duplication(design_unit, self.modules[design_unit.name].source_file.name)
                self.modules[design_unit.name] = design_unit
            elif design_unit.unit_type == "package":
                if design_unit.name in self.verilog_packages:
                    self._warning_on_duplication(
                        design_unit,
                        self.verilog_packages[design_unit.name].source_file.name,
                    )
                self.verilog_packages[design_unit.name] = design_unit

    def get_entities(self):
        """
        Return a list of all entites in the design with their generic names and architecture names
        """
        entities = []
        for entity in self._entities.values():
            entities.append(entity)
        return entities

    def get_modules(self):
        """
        Return a list of all modules in the design
        """
        return list(self.modules.values())

    def get_package_body(self, name):
        return self._package_bodies[name]

    def has_entity(self, name):
        """
        Return true if entity with 'name' is in library
        """
        return name in self._entities

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self.name == other.name

        return False

    def __lt__(self, other):
        return self.name < other.name

    def __hash__(self):
        return hash(self.name)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
Contains Configuration class which contains configuration of a test run
"""

import logging
import inspect
from pathlib import Path
from copy import copy
from vunit.sim_if.factory import SIMULATOR_FACTORY


LOGGER = logging.getLogger(__name__)

# Name of default configuration
DEFAULT_NAME = None


class AttributeException(Exception):
    pass


class Configuration(object):  # pylint: disable=too-many-instance-attributes
    """
    Represents a configuration of a test bench
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        name,
        design_unit,
        generics=None,
        sim_options=None,
        pre_config=None,
        post_check=None,
        attributes=None,
    ):
        self.name = name
        self._design_unit = design_unit
        self.generics = {} if generics is None else generics
        self.sim_options = {} if sim_options is None else sim_options
        self.attributes = {} if attributes is None else attributes

        self.tb_path = str(Path(design_unit.original_file_name).parent)

        # Fill in tb_path generic with location of test bench
        if "tb_path" in design_unit.generic_names:
            self.generics["tb_path"] = str(self.tb_path.replace("\\", "/")) + "/"

        self.pre_config = pre_config
        self.post_check = post_check

    def copy(self):
        return Configuration(
            name=self.name,
            design_unit=self._design_unit,
            generics=self.generics.copy(),
            sim_options=self.sim_options.copy(),
            pre_config=self.pre_config,
            post_check=self.post_check,
            attributes=self.attributes.copy(),
        )

    @property
    def is_default(self):
        return self.name is DEFAULT_NAME

    @property
    def generic_names(self):
        return self._design_unit.generic_names

    @property
    def entity_name(self):
        return self._design_unit.name

    @property
    def design_unit_name(self):
        return self._design_unit.name

    @property
    def library_name(self):
        return self._design_unit.library_name

    @property
    def architecture_name(self):  # pylint: disable=missing-docstring
        if self._design_unit.is_entity:
            return next(iter(self._design_unit.architecture_names))

        return None

    def set_attribute(self, name, value):
        """
        Set attribute
        """
        if name.startswith("."):
            self.attributes[name] = value
        else:
            raise AttributeException

    def set_generic(self, name, value):
        """
        Set generic
        """
        if name not in self._design_unit.generic_names:
            LOGGER.warning(
                "Generic '%s' set to value '%s' not found in %s '%s.%s'. Possible values are [%s]",
                name,
                value,
                "entity" if self._design_unit.is_entity else "module",
                self._design_unit.library_name,
                self._design_unit.name,
                ", ".join(str(gname) for gname in self._design_unit.generic_names),
            )
        else:
            self.generics[name] = value

    def set_sim_option(self, name, value):
        """
        Set sim option
        """
        SIMULATOR_FACTORY.check_sim_option(name, value)
        self.sim_options[name] = copy(value)

    @property
    def vhdl_assert_stop_level(self):
        """
        Return the VHDL assert stop level to use with the simulator
        """
        if "vhdl_assert_stop_level" in self.sim_options:
            level = self.sim_options.get("vhdl_assert_stop_level")
        else:
            level = "error"

        return level

    def call_pre_config(self, output_path, simulator_output_path):
        """
        Call pre_config if available. Setting optional output_path
        """
        if self.pre_config is None:
            return True

        args = inspect.getargspec(self.pre_config).args  # pylint: disable=deprecated-method

        kwargs = {
            "output_path": output_path,
            "simulator_output_path": simulator_output_path,
        }

        for argname in list(kwargs.keys()):
            if argname not in args:
                del kwargs[argname]

        return self.pre_config(**kwargs) is True

    def call_post_check(self, output_path, read_output):
        """
        Call post_check if available. Setting optional output_path
        """
        if self.post_check is None:
            return True

        args = inspect.getargspec(self.post_check).args  # pylint: disable=deprecated-method

        kwargs = {"output_path": lambda: output_path, "output": read_output}

        for argname, provider in list(kwargs.items()):
            if argname not in args:
                del kwargs[argname]
            else:
                kwargs[argname] = provider()

        return self.post_check(**kwargs) is True


class ConfigurationVisitor(object):
    """
    An interface to visit simulation run configurations
    """

    def _check_enabled(self):
        pass

    @staticmethod
    def get_configuration_dicts():
        raise NotImplementedError

    def set_attribute(self, name, value):
        """
        Set attribute
        """
        self._check_enabled()
        for configs in self.get_configuration_dicts():
            for config in configs.values():
                config.set_attribute(name, value)

    def set_generic(self, name, value):
        """
        Set generic
        """
        self._check_enabled()
        for configs in self.get_configuration_dicts():
            for config in configs.values():
                config.set_generic(name, value)

    def set_sim_option(self, name, value, overwrite=True):
        """
        Set sim option

        :param overwrite: To overwrite the option or append to the existing value
        """
        self._check_enabled()
        for configs in self.get_configuration_dicts():
            for config in configs.values():
                if not overwrite:
                    config.set_sim_option(name, config.sim_options.get(name, []) + value)
                    continue
                config.set_sim_option(name, value)

    def set_pre_config(self, value):
        """
        Set pre_config function
        """
        self._check_enabled()
        for configs in self.get_configuration_dicts():
            for config in configs.values():
                config.pre_config = value

    def set_post_check(self, value):
        """
        Set post_check function
        """
        self._check_enabled()
        for configs in self.get_configuration_dicts():
            for config in configs.values():
                config.post_check = value

    def add_config(  # pylint: disable=too-many-arguments
        self,
        name,
        generics=None,
        pre_config=None,
        post_check=None,
        sim_options=None,
        attributes=None,
    ):
        """
        Add a configuration copying unset fields from the default configuration:
        """
        self._check_enabled()

        if name in (DEFAULT_NAME, ""):
            raise ValueError(f"Illegal configuration name {name!r}. Must be non-empty string")

        for configs in self.get_configuration_dicts():
            if name in configs:
                raise RuntimeError(f"Configuration name {name!s} already defined")

            # Copy default configuration
            config = configs[DEFAULT_NAME].copy()
            config.name = name

            if pre_config is not None:
                config.pre_config = pre_config

            if post_check is not None:
                config.post_check = post_check

            if generics is not None:
                config.generics.update(generics)

            if sim_options is not None:
                config.sim_options.update(sim_options)

            if attributes is not None:
                for attribute in attributes:
                    if not attribute.startswith("."):
                        raise AttributeException
                config.attributes.update(attributes)

            configs[config.name] = config



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
Test the ActiveHDL interface
"""


import unittest
from pathlib import Path
import os
from shutil import rmtree
from unittest import mock
from vunit.sim_if.activehdl import ActiveHDLInterface, VersionConsumer, Version
from vunit.project import Project
from vunit.ostools import renew_path, write_file
from vunit.vhdl_standard import VHDL


class MockProcess(object):
    def __init__(self, args, cwd=None, env=None, version_line=None):
        self.args = args
        self.cwd = cwd
        self.env = env
        self.version_line = version_line

    def consume_output(self, callback):
        callback(self.version_line)


class MockProcessVersionWithoutPackageGenerics(MockProcess):
    def __init__(self, args, cwd=None, env=None):
        MockProcess.__init__(self, args, cwd, env, "10.0.12.6914")


class MockProcessVersionWithPackageGenerics(MockProcess):
    def __init__(self, args, cwd=None, env=None):
        MockProcess.__init__(self, args, cwd, env, "10.5a.12.6914")


class TestActiveHDLInterface(unittest.TestCase):
    """
    Test the ActiveHDL interface
    """

    @mock.patch("vunit.sim_if.check_output", autospec=True, return_value="")
    @mock.patch("vunit.sim_if.activehdl.Process", autospec=True)
    def test_compile_project_vhdl_2008(self, process, check_output):
        simif = ActiveHDLInterface(prefix="prefix", output_path=self.output_path)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.vhd", "")
        project.add_source_file("file.vhd", "lib", file_type="vhdl", vhdl_standard=VHDL.standard("2008"))
        simif.compile_project(project)
        process.assert_any_call(
            [str(Path("prefix") / "vlib"), "lib", "lib_path"],
            cwd=self.output_path,
            env=simif.get_env(),
        )
        process.assert_called_with(
            [str(Path("prefix") / "vmap"), "lib", "lib_path"],
            cwd=self.output_path,
            env=simif.get_env(),
        )
        check_output.assert_called_once_with(
            [
                str(Path("prefix") / "vcom"),
                "-quiet",
                "-j",
                self.output_path,
                "-2008",
                "-work",
                "lib",
                "file.vhd",
            ],
            env=simif.get_env(),
        )

    @mock.patch("vunit.sim_if.check_output", autospec=True, return_value="")
    @mock.patch("vunit.sim_if.activehdl.Process", autospec=True)
    def test_compile_project_vhdl_2002(self, process, check_output):
        simif = ActiveHDLInterface(prefix="prefix", output_path=self.output_path)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.vhd", "")
        project.add_source_file("file.vhd", "lib", file_type="vhdl", vhdl_standard=VHDL.standard("2002"))
        simif.compile_project(project)
        process.assert_any_call(
            [str(Path("prefix") / "vlib"), "lib", "lib_path"],
            cwd=self.output_path,
            env=simif.get_env(),
        )
        process.assert_called_with(
            [str(Path("prefix") / "vmap"), "lib", "lib_path"],
            cwd=self.output_path,
            env=simif.get_env(),
        )
        check_output.assert_called_once_with(
            [
                str(Path("prefix") / "vcom"),
                "-quiet",
                "-j",
                self.output_path,
                "-2002",
                "-work",
                "lib",
                "file.vhd",
            ],
            env=simif.get_env(),
        )

    @mock.patch("vunit.sim_if.check_output", autospec=True, return_value="")
    @mock.patch("vunit.sim_if.activehdl.Process", autospec=True)
    def test_compile_project_vhdl_93(self, process, check_output):
        simif = ActiveHDLInterface(prefix="prefix", output_path=self.output_path)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.vhd", "")
        project.add_source_file("file.vhd", "lib", file_type="vhdl", vhdl_standard=VHDL.standard("93"))
        simif.compile_project(project)
        process.assert_any_call(
            [str(Path("prefix") / "vlib"), "lib", "lib_path"],
            cwd=self.output_path,
            env=simif.get_env(),
        )
        process.assert_called_with(
            [str(Path("prefix") / "vmap"), "lib", "lib_path"],
            cwd=self.output_path,
            env=simif.get_env(),
        )
        check_output.assert_called_once_with(
            [
                str(Path("prefix") / "vcom"),
                "-quiet",
                "-j",
                self.output_path,
                "-93",
                "-work",
                "lib",
                "file.vhd",
            ],
            env=simif.get_env(),
        )

    @mock.patch("vunit.sim_if.check_output", autospec=True, return_value="")
    @mock.patch("vunit.sim_if.activehdl.Process", autospec=True)
    def test_compile_project_vhdl_extra_flags(self, process, check_output):
        simif = ActiveHDLInterface(prefix="prefix", output_path=self.output_path)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.vhd", "")
        source_file = project.add_source_file("file.vhd", "lib", file_type="vhdl")
        source_file.set_compile_option("activehdl.vcom_flags", ["custom", "flags"])
        simif.compile_project(project)
        process.assert_any_call(
            [str(Path("prefix") / "vlib"), "lib", "lib_path"],
            cwd=self.output_path,
            env=simif.get_env(),
        )
        process.assert_called_with(
            [str(Path("prefix") / "vmap"), "lib", "lib_path"],
            cwd=self.output_path,
            env=simif.get_env(),
        )
        check_output.assert_called_once_with(
            [
                str(Path("prefix") / "vcom"),
                "-quiet",
                "-j",
                self.output_path,
                "custom",
                "flags",
                "-2008",
                "-work",
                "lib",
                "file.vhd",
            ],
            env=simif.get_env(),
        )

    @mock.patch("vunit.sim_if.check_output", autospec=True, return_value="")
    @mock.patch("vunit.sim_if.activehdl.Process", autospec=True)
    def test_compile_project_verilog(self, process, check_output):
        library_cfg = str(Path(self.output_path) / "library.cfg")
        simif = ActiveHDLInterface(prefix="prefix", output_path=self.output_path)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.v", "")
        project.add_source_file("file.v", "lib", file_type="verilog")
        simif.compile_project(project)
        process.assert_any_call(
            [str(Path("prefix") / "vlib"), "lib", "lib_path"],
            cwd=self.output_path,
            env=simif.get_env(),
        )
        process.assert_called_with(
            [str(Path("prefix") / "vmap"), "lib", "lib_path"],
            cwd=self.output_path,
            env=simif.get_env(),
        )
        check_output.assert_called_once_with(
            [
                str(Path("prefix") / "vlog"),
                "-quiet",
                "-lc",
                library_cfg,
                "-work",
                "lib",
                "file.v",
                "-l",
                "lib",
            ],
            env=simif.get_env(),
        )

    @mock.patch("vunit.sim_if.check_output", autospec=True, return_value="")
    @mock.patch("vunit.sim_if.activehdl.Process", autospec=True)
    def test_compile_project_system_verilog(self, process, check_output):
        library_cfg = str(Path(self.output_path) / "library.cfg")
        simif = ActiveHDLInterface(prefix="prefix", output_path=self.output_path)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.sv", "")
        project.add_source_file("file.sv", "lib", file_type="systemverilog")
        simif.compile_project(project)
        process.assert_any_call(
            [str(Path("prefix") / "vlib"), "lib", "lib_path"],
            cwd=self.output_path,
            env=simif.get_env(),
        )
        process.assert_called_with(
            [str(Path("prefix") / "vmap"), "lib", "lib_path"],
            cwd=self.output_path,
            env=simif.get_env(),
        )
        check_output.assert_called_once_with(
            [
                str(Path("prefix") / "vlog"),
                "-quiet",
                "-lc",
                library_cfg,
                "-work",
                "lib",
                "file.sv",
                "-l",
                "lib",
            ],
            env=simif.get_env(),
        )

    @mock.patch("vunit.sim_if.check_output", autospec=True, return_value="")
    @mock.patch("vunit.sim_if.activehdl.Process", autospec=True)
    def test_compile_project_verilog_extra_flags(self, process, check_output):
        library_cfg = str(Path(self.output_path) / "library.cfg")
        simif = ActiveHDLInterface(prefix="prefix", output_path=self.output_path)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.v", "")
        source_file = project.add_source_file("file.v", "lib", file_type="verilog")
        source_file.set_compile_option("activehdl.vlog_flags", ["custom", "flags"])
        simif.compile_project(project)
        process.assert_any_call(
            [str(Path("prefix") / "vlib"), "lib", "lib_path"],
            cwd=self.output_path,
            env=simif.get_env(),
        )
        process.assert_called_with(
            [str(Path("prefix") / "vmap"), "lib", "lib_path"],
            cwd=self.output_path,
            env=simif.get_env(),
        )
        check_output.assert_called_once_with(
            [
                str(Path("prefix") / "vlog"),
                "-quiet",
                "-lc",
                library_cfg,
                "custom",
                "flags",
                "-work",
                "lib",
                "file.v",
                "-l",
                "lib",
            ],
            env=simif.get_env(),
        )

    @mock.patch("vunit.sim_if.check_output", autospec=True, return_value="")
    @mock.patch("vunit.sim_if.activehdl.Process", autospec=True)
    def test_compile_project_verilog_include(self, process, check_output):
        library_cfg = str(Path(self.output_path) / "library.cfg")
        simif = ActiveHDLInterface(prefix="prefix", output_path=self.output_path)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.v", "")
        project.add_source_file("file.v", "lib", file_type="verilog", include_dirs=["include"])
        simif.compile_project(project)
        process.assert_any_call(
            [str(Path("prefix") / "vlib"), "lib", "lib_path"],
            cwd=self.output_path,
            env=simif.get_env(),
        )
        process.assert_called_with(
            [str(Path("prefix") / "vmap"), "lib", "lib_path"],
            cwd=self.output_path,
            env=simif.get_env(),
        )
        check_output.assert_called_once_with(
            [
                str(Path("prefix") / "vlog"),
                "-quiet",
                "-lc",
                library_cfg,
                "-work",
                "lib",
                "file.v",
                "-l",
                "lib",
                "+incdir+include",
            ],
            env=simif.get_env(),
        )

    @mock.patch("vunit.sim_if.check_output", autospec=True, return_value="")
    @mock.patch("vunit.sim_if.activehdl.Process", autospec=True)
    def test_compile_project_verilog_define(self, process, check_output):
        library_cfg = str(Path(self.output_path) / "library.cfg")
        simif = ActiveHDLInterface(prefix="prefix", output_path=self.output_path)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.v", "")
        project.add_source_file("file.v", "lib", file_type="verilog", defines={"defname": "defval"})
        simif.compile_project(project)
        process.assert_any_call(
            [str(Path("prefix") / "vlib"), "lib", "lib_path"],
            cwd=self.output_path,
            env=simif.get_env(),
        )
        process.assert_called_with(
            [str(Path("prefix") / "vmap"), "lib", "lib_path"],
            cwd=self.output_path,
            env=simif.get_env(),
        )
        check_output.assert_called_once_with(
            [
                str(Path("prefix") / "vlog"),
                "-quiet",
                "-lc",
                library_cfg,
                "-work",
                "lib",
                "file.v",
                "-l",
                "lib",
                "+define+defname=defval",
            ],
            env=simif.get_env(),
        )

    @mock.patch("vunit.sim_if.activehdl.ActiveHDLInterface.find_prefix")
    @mock.patch("vunit.sim_if.activehdl.Process", new=MockProcessVersionWithPackageGenerics)
    def test_supports_vhdl_package_generics_true(self, find_prefix):
        find_prefix.return_value = ""
        simif = ActiveHDLInterface(prefix="prefix", output_path=self.output_path)
        self.assertTrue(simif.supports_vhdl_package_generics())

    @mock.patch("vunit.sim_if.activehdl.ActiveHDLInterface.find_prefix")
    @mock.patch("vunit.sim_if.activehdl.Process", new=MockProcessVersionWithoutPackageGenerics)
    def test_supports_vhdl_package_generics_false(self, find_prefix):
        find_prefix.return_value = ""
        simif = ActiveHDLInterface(prefix="prefix", output_path=self.output_path)
        self.assertFalse(simif.supports_vhdl_package_generics())

    def setUp(self):
        self.output_path = str(Path(__file__).parent / "test_activehdl_out")
        renew_path(self.output_path)
        self.project = Project()
        self.cwd = os.getcwd()
        os.chdir(self.output_path)

    def tearDown(self):
        os.chdir(self.cwd)
        if Path(self.output_path).exists():
            rmtree(self.output_path)


class TestVersionConsumer(unittest.TestCase):
    """
    Test the VersionConsumer class
    """

    def _assert_version_correct(self, version_line, expected_major, expected_minor, expected_minor_letter):
        """
        Assertion function used by tests in this class
        """
        consumer = VersionConsumer()
        consumer(version_line)
        self.assertEqual(
            consumer.version,
            Version(expected_major, expected_minor, expected_minor_letter),
        )

    def test_vendor_version_without_letters(self):
        self._assert_version_correct(
            "Aldec, Inc. VHDL compiler version 10.5.216.6767 built for Windows on January 20, 2018.",
            10,
            5,
            "",
        )

    def test_vendor_version_with_letters(self):
        self._assert_version_correct(
            "Aldec, Inc. VHDL compiler version 10.5a.12.6914 built for Windows on June 06, 2018.",
            10,
            5,
            "a",
        )


class TestVersion(unittest.TestCase):
    """
    Test the Version class.
    Test cases = (assert true, assert false) x (with letters, without letters, mixed) x number_of_operations
    Where number_of_operations = 1 for <, and 2 for <=, etc
    """

    high_version_letter = Version(10, 5, "b")
    low_version_letter = Version(10, 5, "a")
    high_version_no_letter = Version(10, 6)
    low_version_no_letter = Version(10, 5)
    high_version_letter_for_mixed = Version(10, 6, "")

    def test_lt(self):
        # Test with letters
        self.assertTrue(TestVersion.low_version_letter < TestVersion.high_version_letter)
        self.assertFalse(TestVersion.high_version_letter < TestVersion.low_version_letter)
        # Test without letters
        self.assertTrue(TestVersion.low_version_no_letter < TestVersion.high_version_no_letter)
        self.assertFalse(TestVersion.high_version_no_letter < TestVersion.low_version_no_letter)
        # Both
        self.assertTrue(TestVersion.low_version_letter < TestVersion.high_version_no_letter)
        self.assertFalse(TestVersion.high_version_letter < TestVersion.low_version_no_letter)

    def test_le(self):
        # Test equal
        # Test with letters
        self.assertTrue(TestVersion.high_version_letter <= TestVersion.high_version_letter)
        self.assertFalse(TestVersion.high_version_letter <= TestVersion.low_version_letter)
        # Test without letters
        self.assertTrue(TestVersion.high_version_no_letter <= TestVersion.high_version_no_letter)
        self.assertFalse(TestVersion.high_version_no_letter <= TestVersion.low_version_no_letter)
        # Both
        self.assertTrue(TestVersion.high_version_letter <= TestVersion.high_version_no_letter)
        self.assertFalse(TestVersion.high_version_letter <= TestVersion.low_version_no_letter)

        # Test less than
        # Test with letters
        self.assertTrue(TestVersion.low_version_letter <= TestVersion.high_version_letter)
        self.assertFalse(TestVersion.high_version_letter <= TestVersion.low_version_letter)
        # Test without letters
        self.assertTrue(TestVersion.low_version_no_letter <= TestVersion.high_version_no_letter)
        self.assertFalse(TestVersion.high_version_no_letter <= TestVersion.low_version_no_letter)
        # Both
        self.assertTrue(TestVersion.low_version_letter <= TestVersion.high_version_no_letter)
        self.assertFalse(TestVersion.high_version_letter <= TestVersion.low_version_no_letter)

    def test_gt(self):
        # Test with letters
        self.assertTrue(TestVersion.high_version_letter > TestVersion.low_version_letter)
        self.assertFalse(TestVersion.low_version_letter > TestVersion.high_version_letter)
        # Test without letters
        self.assertTrue(TestVersion.high_version_no_letter > TestVersion.low_version_no_letter)
        self.assertFalse(TestVersion.low_version_no_letter > TestVersion.high_version_no_letter)
        # Both
        self.assertTrue(TestVersion.high_version_letter > TestVersion.low_version_no_letter)
        self.assertFalse(TestVersion.low_version_letter > TestVersion.high_version_no_letter)

    def test_ge(self):
        # Test equal
        # Test with letters
        self.assertTrue(TestVersion.high_version_letter >= TestVersion.high_version_letter)
        self.assertFalse(TestVersion.low_version_letter >= TestVersion.high_version_letter)
        # Test without letters
        self.assertTrue(TestVersion.high_version_no_letter >= TestVersion.high_version_no_letter)
        self.assertFalse(TestVersion.low_version_no_letter >= TestVersion.high_version_no_letter)
        # Both
        self.assertTrue(TestVersion.high_version_letter_for_mixed >= TestVersion.high_version_no_letter)
        self.assertFalse(TestVersion.low_version_letter >= TestVersion.high_version_no_letter)

        # Test greater than
        # Test with letters
        self.assertTrue(TestVersion.high_version_letter >= TestVersion.low_version_letter)
        self.assertFalse(TestVersion.low_version_letter >= TestVersion.high_version_letter)
        # Test without letters
        self.assertTrue(TestVersion.high_version_no_letter >= TestVersion.low_version_no_letter)
        self.assertFalse(TestVersion.low_version_no_letter >= TestVersion.high_version_no_letter)
        # Both
        self.assertTrue(TestVersion.high_version_letter >= TestVersion.low_version_no_letter)
        self.assertFalse(TestVersion.low_version_letter >= TestVersion.high_version_no_letter)

    def test_eq(self):
        # Test with letters
        self.assertTrue(TestVersion.high_version_letter == TestVersion.high_version_letter)
        self.assertFalse(TestVersion.high_version_letter == TestVersion.low_version_letter)
        # Test without letters
        self.assertTrue(TestVersion.high_version_no_letter == TestVersion.high_version_no_letter)
        self.assertFalse(TestVersion.high_version_no_letter == TestVersion.low_version_no_letter)
        # Both
        self.assertTrue(TestVersion.high_version_letter_for_mixed == TestVersion.high_version_no_letter)
        self.assertFalse(TestVersion.high_version_letter == TestVersion.low_version_no_letter)

    def test_ne(self):
        # Test with letters
        self.assertTrue(TestVersion.high_version_letter != TestVersion.low_version_letter)
        self.assertFalse(TestVersion.high_version_letter != TestVersion.high_version_letter)
        # Test without letters
        self.assertTrue(TestVersion.high_version_no_letter != TestVersion.low_version_no_letter)
        self.assertFalse(TestVersion.high_version_no_letter != TestVersion.high_version_no_letter)
        # Both
        self.assertTrue(TestVersion.high_version_letter != TestVersion.low_version_no_letter)
        self.assertFalse(TestVersion.high_version_letter_for_mixed != TestVersion.high_version_no_letter)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
Functions to add builtin VHDL code to a project for compilation
"""

from pathlib import Path
from glob import glob
from warnings import warn
from vunit.vhdl_standard import VHDL, VHDLStandard
from vunit.ui.common import get_checked_file_names_from_globs

VHDL_PATH = (Path(__file__).parent / "vhdl").resolve()
VERILOG_PATH = (Path(__file__).parent / "verilog").resolve()


class Builtins(object):
    """
    Manage VUnit builtins and their dependencies
    """

    def __init__(self, vunit_obj, vhdl_standard: VHDLStandard, simulator_class):
        self._vunit_obj = vunit_obj
        self._vunit_lib = vunit_obj.add_library("vunit_lib")
        self._vhdl_standard = vhdl_standard
        self._simulator_class = simulator_class
        self._builtins_adder = BuiltinsAdder()

        def add(name, deps=tuple()):
            self._builtins_adder.add_type(name, getattr(self, f"_add_{name!s}"), deps)

        add("array_util")
        add("com")
        add("verification_components", ["com", "osvvm"])
        add("osvvm")
        add("random", ["osvvm"])
        add("json4vhdl")

    def add(self, name, args=None):
        self._builtins_adder.add(name, args)

    def _add_files(self, pattern=None, allow_empty=True):
        """
        Add files with naming convention to indicate which standard is supported
        """
        supports_context = self._simulator_class.supports_vhdl_contexts() and self._vhdl_standard.supports_context

        for file_name in get_checked_file_names_from_globs(pattern, allow_empty):
            base_file_name = Path(file_name).name

            standards = set()
            for standard in VHDL.STANDARDS:
                standard_name = str(standard)
                if standard_name + "p" in base_file_name:
                    standards.update(standard.and_later)
                elif standard_name + "m" in base_file_name:
                    standards.update(standard.and_earlier)
                elif standard_name in base_file_name:
                    standards.add(standard)

            if standards and self._vhdl_standard not in standards:
                continue

            if not supports_context and file_name.endswith("_context.vhd"):
                continue

            self._vunit_lib.add_source_file(file_name)

    def _add_data_types(self, external=None):
        """
        Add data types packages (sources corresponding to VHPIDIRECT arrays, or their placeholders)

        :param external: struct to provide bridges for the external VHDL API.
                         {
                             'string': ['path/to/custom/file'],
                             'integer': ['path/to/custom/file']
                         }.
        """
        self._add_files(VHDL_PATH / "data_types" / "src" / "*.vhd")

        for key in ["string", "integer_vector"]:
            self._add_files(
                pattern=str(VHDL_PATH / "data_types" / "src" / "api" / f"external_{key!s}_pkg.vhd")
                if external is None or key not in external or not external[key] or external[key] is True
                else external[key],
                allow_empty=False,
            )

    def _add_array_util(self):
        """
        Add array utility
        """
        if not self._vhdl_standard >= VHDL.STD_2008:
            raise RuntimeError("Array util only supports vhdl 2008 and later")

        arr_deprecation_note = (
            "'array_t' is deprecated and it will removed in future releases; use 'integer_array_t' instead"
        )
        warn(arr_deprecation_note, Warning)

        self._vunit_lib.add_source_files(VHDL_PATH / "array" / "src" / "*.vhd")

    def _add_random(self):
        """
        Add random pkg
        """
        if not self._vhdl_standard >= VHDL.STD_2008:
            raise RuntimeError("Random only supports vhdl 2008 and later")

        self._vunit_lib.add_source_files(VHDL_PATH / "random" / "src" / "*.vhd")

    def _add_com(self):
        """
        Add com library
        """
        if not self._vhdl_standard >= VHDL.STD_2008:
            raise RuntimeError("Communication package only supports vhdl 2008 and later")

        self._add_files(VHDL_PATH / "com" / "src" / "*.vhd")

    def _add_verification_components(self):
        """
        Add verification component library
        """
        if not self._vhdl_standard >= VHDL.STD_2008:
            raise RuntimeError("Verification component library only supports vhdl 2008 and later")
        self._add_files(VHDL_PATH / "verification_components" / "src" / "*.vhd")

    def _add_osvvm(self):
        """
        Add osvvm library
        """
        library_name = "osvvm"

        try:
            library = self._vunit_obj.library(library_name)
        except KeyError:
            library = self._vunit_obj.add_library(library_name)

        simulator_coverage_api = self._simulator_class.get_osvvm_coverage_api()
        supports_vhdl_package_generics = self._simulator_class.supports_vhdl_package_generics()

        if not osvvm_is_installed():
            raise RuntimeError(
                """
Found no OSVVM VHDL files. Did you forget to run

git submodule update --init --recursive

in your VUnit Git repository? You have to do this first if installing using setup.py."""
            )

        for file_name in glob(str(VHDL_PATH / "osvvm" / "*.vhd")):
            bname = Path(file_name).name

            if (bname == "AlertLogPkg_body_BVUL.vhd") or ("2019" in bname):
                continue

            if ((simulator_coverage_api != "rivierapro") and (bname == "VendorCovApiPkg_Aldec.vhd")) or (
                (simulator_coverage_api == "rivierapro") and (bname == "VendorCovApiPkg.vhd")
            ):
                continue

            if not supports_vhdl_package_generics and (
                bname
                in [
                    "ScoreboardGenericPkg.vhd",
                    "ScoreboardPkg_int.vhd",
                    "ScoreboardPkg_slv.vhd",
                ]
            ):
                continue

            library.add_source_files(file_name, preprocessors=[])

    def _add_json4vhdl(self):
        """
        Add JSON-for-VHDL library
        """
        library_name = "JSON"

        try:
            library = self._vunit_obj.library(library_name)
        except KeyError:
            library = self._vunit_obj.add_library(library_name)

        library.add_source_files(VHDL_PATH / "JSON-for-VHDL" / "src" / "*.vhdl")

    def add_verilog_builtins(self):
        """
        Add Verilog builtins
        """
        self._vunit_lib.add_source_files(VERILOG_PATH / "vunit_pkg.sv")

    def add_vhdl_builtins(self, external=None):
        """
        Add vunit VHDL builtin libraries

        :param external: struct to provide bridges for the external VHDL API.
                         {
                             'string': ['path/to/custom/file'],
                             'integer': ['path/to/custom/file']
                         }.
        """
        self._add_data_types(external=external)
        self._add_files(VHDL_PATH / "*.vhd")
        for path in (
            "core",
            "logging",
            "string_ops",
            "check",
            "dictionary",
            "run",
            "path",
        ):
            self._add_files(VHDL_PATH / path / "src" / "*.vhd")


def osvvm_is_installed():
    """
    Checks if OSVVM is installed within the VUnit directory structure
    """
    return len(glob(str(VHDL_PATH / "osvvm" / "*.vhd"))) != 0


def add_verilog_include_dir(include_dirs):
    """
    Add VUnit Verilog include directory
    """
    return [str(VERILOG_PATH / "include")] + include_dirs


class BuiltinsAdder(object):
    """
    Class to manage adding of builtins with dependencies
    """

    def __init__(self):
        self._already_added = {}
        self._types = {}

    def add_type(self, name, function, dependencies=tuple()):
        self._types[name] = (function, dependencies)

    def add(self, name, args=None):
        """
        Add builtin with arguments
        """
        args = {} if args is None else args

        if not self._add_check(name, args):
            function, dependencies = self._types[name]
            for dep_name in dependencies:
                self.add(dep_name)
            function(**args)

    def _add_check(self, name, args=None):
        """
        Check if this package has already been added,
        if it has already been added it must use the same parameters

        @returns False if not yet added
        """
        if name not in self._already_added:
            self._already_added[name] = args
            return False

        old_args = self._already_added[name]
        if args != old_args:
            raise RuntimeError(
                f"Optional builtin {name!r} added with arguments {args!r} "
                f"has already been added with arguments {old_args!r}"
            )
        return True



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
UI class Library
"""

from pathlib import Path
from fnmatch import fnmatch
from typing import Optional
from ..vhdl_standard import VHDL, VHDLStandard
from ..project import Project
from ..source_file import file_type_of, FILE_TYPES, VERILOG_FILE_TYPES
from ..builtins import add_verilog_include_dir
from .common import check_not_empty, get_checked_file_names_from_globs
from .source import SourceFile, SourceFileList
from .testbench import TestBench
from .packagefacade import PackageFacade


class Library(object):
    """
    User interface of a library
    """

    def __init__(self, library_name, parent, project: Project, test_bench_list):
        self._library_name = library_name
        self._parent = parent
        self._project = project
        self._test_bench_list = test_bench_list

    @property
    def name(self):
        """
        The name of the library
        """
        return self._library_name

    def set_generic(self, name, value, allow_empty=False):
        """
        Set a value of generic within all |configurations| of test benches and tests this library

        :param name: The name of the generic
        :param value: The value of the generic
        :param allow_empty: To disable an error when no test benches were found

        :example:

        .. code-block:: python

           lib.set_generic("data_width", 16)

        .. note::
           Only affects test benches added *before* the generic is set.
        """
        for test_bench in self.get_test_benches(allow_empty=allow_empty):
            test_bench.set_generic(name.lower(), value)

    def set_parameter(self, name, value, allow_empty=False):
        """
        Set a value of parameter within all |configurations| of test benches and tests this library

        :param name: The name of the parameter
        :param value: The value of the parameter
        :param allow_empty: To disable an error when no test benches were found

        :example:

        .. code-block:: python

           lib.set_parameter("data_width", 16)

        .. note::
           Only affects test benches added *before* the parameter is set.
        """
        for test_bench in self.get_test_benches(allow_empty=allow_empty):
            test_bench.set_generic(name, value)

    def set_sim_option(self, name, value, allow_empty=False, overwrite=True):
        """
        Set simulation option within all |configurations| of test benches and tests this library

        :param name: |simulation_options|
        :param value: The value of the simulation option
        :param allow_empty: To disable an error when no test benches were found
        :param overwrite: To overwrite the option or append to the existing value

        :example:

        .. code-block:: python

           lib.set_sim_option("ghdl.a_flags", ["--no-vital-checks"])

        .. note::
           Only affects test benches added *before* the option is set.
        """
        for test_bench in self.get_test_benches(allow_empty=allow_empty):
            test_bench.set_sim_option(name, value, overwrite)

    def set_compile_option(self, name, value, allow_empty=False):
        """
        Set compile option for all files within the library

        :param name: |compile_option|
        :param value: The value of the compile option
        :param allow_empty: To disable an error when no source files were found

        :example:

        .. code-block:: python

           lib.set_compile_option("ghdl.a_flags", ["--no-vital-checks"])


        .. note::
           Only affects files added *before* the option is set.
        """
        for source_file in self.get_source_files(allow_empty=allow_empty):
            source_file.set_compile_option(name, value)

    def add_compile_option(self, name, value, allow_empty=False):
        """
        Add compile option to all files within the library

        :param name: |compile_option|
        :param value: The value of the compile option
        :param allow_empty: To disable an error when no source files were found

        .. note::
           Only affects files added *before* the option is set.
        """
        for source_file in self.get_source_files(allow_empty=allow_empty):
            source_file.add_compile_option(name, value)

    def get_source_file(self, file_name):
        """
        Get a source file within this library

        :param file_name: The name of the file as a relative or absolute path

        :returns: A :class:`.SourceFile` object
        """
        return self._parent.get_source_file(file_name, self._library_name)

    def get_source_files(self, pattern="*", allow_empty=False):
        """
        Get a list of source files within this libary

        :param pattern: A wildcard pattern matching either an absolute or relative path
        :param allow_empty: To disable an error if no files matched the pattern
        :returns: A :class:`.SourceFileList` object
        """
        return self._parent.get_source_files(pattern, self._library_name, allow_empty)

    def add_source_files(  # pylint: disable=too-many-arguments
        self,
        pattern,
        preprocessors=None,
        include_dirs=None,
        defines=None,
        allow_empty=False,
        vhdl_standard: Optional[str] = None,
        no_parse=False,
        file_type=None,
    ):
        """
        Add source files matching wildcard pattern to library

        :param pattern: A wildcard pattern match the files to add
        :param include_dirs: A list of include directories
        :param defines: A dictionary containing Verilog defines to be set
        :param allow_empty: To disable an error if no files matched the pattern
        :param vhdl_standard: The VHDL standard used to compile files, if None library default is used
        :param no_parse: Do not parse file(s) for dependency or test scanning purposes
        :param file_type: The type of the file; ``"vhdl"``, ``"verilog"``  or ``"systemverilog"``.
                          Auto-detected by default when set to ``None``.
        :returns: A list of files (:class:`.SourceFileList`) which were added

        :example:

        .. code-block:: python

           library.add_source_files("*.vhd")

        """
        return SourceFileList(
            source_files=[
                self.add_source_file(
                    file_name,
                    preprocessors,
                    include_dirs,
                    defines,
                    vhdl_standard,
                    no_parse=no_parse,
                    file_type=file_type,
                )
                for file_name in get_checked_file_names_from_globs(pattern, allow_empty)
            ]
        )

    def add_source_file(  # pylint: disable=too-many-arguments
        self,
        file_name,
        preprocessors=None,
        include_dirs=None,
        defines=None,
        vhdl_standard: Optional[str] = None,
        no_parse=False,
        file_type=None,
    ):
        """
        Add source file to library

        :param file_name: The name of the file
        :param include_dirs: A list of include directories
        :param defines: A dictionary containing Verilog defines to be set
        :param vhdl_standard: The VHDL standard used to compile this file, if None library default is used
        :param no_parse: Do not parse file for dependency or test scanning purposes
        :param file_type: The type of the file; ``"vhdl"``, ``"verilog"``  or ``"systemverilog"``.
                          Auto-detected by default when set to ``None``.
        :returns: The :class:`.SourceFile` which was added

        :example:

        .. code-block:: python

           library.add_source_file("file.vhd")

        """
        file_name = Path(file_name).resolve()

        if file_type is None:
            file_type = file_type_of(file_name)
        elif file_type not in FILE_TYPES:
            raise ValueError(f"file_type {file_type!r} not in {FILE_TYPES!r}")

        if file_type in VERILOG_FILE_TYPES:
            include_dirs = include_dirs if include_dirs is not None else []
            include_dirs = add_verilog_include_dir(include_dirs)

        new_file_name = self._parent._preprocess(  # pylint: disable=protected-access
            self._library_name, file_name, preprocessors
        )

        source_file = self._project.add_source_file(
            new_file_name,
            self._library_name,
            file_type=file_type,
            include_dirs=include_dirs,
            defines=defines,
            vhdl_standard=self._which_vhdl_standard(vhdl_standard),
            no_parse=no_parse,
        )
        # To get correct tb_path generic
        source_file.original_name = file_name

        self._test_bench_list.add_from_source_file(source_file)

        return SourceFile(source_file, self._project, self._parent)

    def package(self, name):
        """
        Get a package within the library
        """
        library = self._project.get_library(self._library_name)
        design_unit = library.primary_design_units.get(name)

        if design_unit is None:
            raise KeyError(name)
        if design_unit.unit_type != "package":
            raise KeyError(name)

        return PackageFacade(self._parent, self._library_name, name, design_unit)

    def entity(self, name):
        """
        Get an entity within the library

        :param name: The name of the entity
        :returns: A :class:`.TestBench` object
        :raises: KeyError
        """
        name = name.lower()
        library = self._project.get_library(self._library_name)
        if not library.has_entity(name):
            raise KeyError(name)

        return self.test_bench(name)

    def module(self, name):
        """
        Get a module within the library

        :param name: The name of the module
        :returns: A :class:`.TestBench` object
        :raises: KeyError
        """
        library = self._project.get_library(self._library_name)
        if name not in library.modules:
            raise KeyError(name)

        return self.test_bench(name)

    def test_bench(self, name):
        """
        Get a test bench within this library

        :param name: The name of the test bench
        :returns: A :class:`.TestBench` object
        :raises: KeyError
        """
        name = name.lower()

        return TestBench(self._test_bench_list.get_test_bench(self._library_name, name), self)

    def get_test_benches(self, pattern="*", allow_empty=False):
        """
        Get a list of test benches

        :param pattern: A wildcard pattern matching the test_bench name
        :param allow_empty: To disable an error when no test benches were found
        :returns: A list of :class:`.TestBench` objects
        """
        results = []
        for test_bench in self._test_bench_list.get_test_benches_in_library(self._library_name):
            if not fnmatch(Path(test_bench.name).resolve(), pattern):
                continue

            results.append(TestBench(test_bench, self))

        return check_not_empty(
            results,
            allow_empty,
            f"No test benches found within library {self._library_name!s}",
        )

    def _which_vhdl_standard(self, vhdl_standard: Optional[str]) -> VHDLStandard:
        """
        Return default vhdl_standard if the argument is None
        The argument is a string from the user
        """
        if vhdl_standard is None:
            return self._project.get_library(self._library_name).vhdl_standard

        return VHDL.standard(vhdl_standard)
